from dataclasses import dataclass
from functools import cached_property

from prompt_toolkit.formatted_text import FormattedText, to_plain_text
from pymysql.cursors import Cursor


@dataclass
class SQLResult:
    preamble: str | None = None
    header: list[str] | str | None = None
    rows: Cursor | list[tuple] | None = None
    postamble: str | None = None
    status: str | FormattedText | None = None
    command: dict[str, str | float] | None = None

    def __str__(self):
        return f"{self.preamble}, {self.header}, {self.rows}, {self.postamble}, {self.status}, {self.command}"

    @cached_property
    def status_plain(self):
        if self.status is None:
            return None
        return to_plain_text(self.status)



from __future__ import annotations

import datetime
import enum
import logging
import re
import ssl
from typing import Any, Generator, Iterable

from prompt_toolkit.formatted_text import FormattedText
import pymysql
from pymysql.connections import Connection
from pymysql.constants import FIELD_TYPE
from pymysql.converters import conversions, convert_date, convert_datetime, convert_time, decoders
from pymysql.cursors import Cursor

from mycli.constants import ER_MUST_CHANGE_PASSWORD
from mycli.packages.special import iocommands
from mycli.packages.special.main import CommandNotFound, execute
from mycli.packages.sqlresult import SQLResult

try:
    import paramiko  # noqa: F401
    import sshtunnel
except ImportError:
    pass

_logger = logging.getLogger(__name__)

FIELD_TYPES = decoders.copy()
FIELD_TYPES.update({FIELD_TYPE.NULL: type(None)})


ERROR_CODE_ACCESS_DENIED = 1045


class ServerSpecies(enum.Enum):
    MySQL = "MySQL"
    MariaDB = "MariaDB"
    Percona = "Percona"
    TiDB = "TiDB"
    Doris = "Doris"
    Unknown = "Unknown"


class ServerInfo:
    def __init__(self, species: ServerSpecies | None, version_str: str) -> None:
        self.species = species
        self.version_str = version_str
        self.version = self.calc_mysql_version_value(version_str)

    @staticmethod
    def calc_mysql_version_value(version_str: str) -> int:
        if not version_str or not isinstance(version_str, str):
            return 0
        try:
            major, minor, patch = version_str.split(".")
        except ValueError:
            return 0
        else:
            return int(major) * 10_000 + int(minor) * 100 + int(patch)

    @classmethod
    def from_version_string(cls, version_string: str) -> ServerInfo:
        if not version_string:
            return cls(ServerSpecies.MySQL, "")

        re_species = (
            (r"(?P<version>[0-9\.]+)-MariaDB", ServerSpecies.MariaDB),
            (r"[0-9\.]*-TiDB-v(?P<version>[0-9\.]+)-?(?P<comment>[a-z0-9\-]*)", ServerSpecies.TiDB),
            (r"(?P<version>[0-9\.]+)[a-z0-9]*-(?P<comment>[0-9]+$)", ServerSpecies.Percona),
            # Also matches plain "X.Y.Z" with no suffix (e.g. Homebrew MySQL).
            (r"(?P<version>[0-9]+\.[0-9]+\.[0-9]+)[a-z0-9]*(-(?P<comment>[A-Za-z0-9_]+))?", ServerSpecies.MySQL),
        )
        for regexp, species in re_species:
            match = re.search(regexp, version_string)
            if match is not None:
                parsed_version = match.group("version")
                detected_species = species
                break
        else:
            detected_species = ServerSpecies.MySQL
            parsed_version = ""

        return cls(detected_species, parsed_version)

    def __str__(self) -> str:
        if self.species:
            return f"{self.species.value} {self.version_str}"
        else:
            return self.version_str


class SQLExecute:
    databases_query = """SHOW DATABASES"""

    tables_query = """SHOW TABLES"""

    show_candidates_query = '''SELECT name from mysql.help_topic WHERE name like "SHOW %"'''

    users_query = """SELECT CONCAT("'", user, "'@'",host,"'") FROM mysql.user"""

    functions_query = '''SELECT ROUTINE_NAME FROM INFORMATION_SCHEMA.ROUTINES
    WHERE ROUTINE_TYPE="FUNCTION" AND ROUTINE_SCHEMA = %s'''

    procedures_query = '''SELECT ROUTINE_NAME FROM INFORMATION_SCHEMA.ROUTINES
    WHERE ROUTINE_TYPE="PROCEDURE" AND ROUTINE_SCHEMA = %s'''

    character_sets_query = '''SHOW CHARACTER SET'''

    collations_query = '''SHOW COLLATION'''

    table_columns_query = """select TABLE_NAME, COLUMN_NAME from information_schema.columns
                                    where table_schema = %s
                                    order by table_name,ordinal_position"""

    enum_values_query = """select TABLE_NAME, COLUMN_NAME, COLUMN_TYPE from information_schema.columns
                                    where table_schema = %s and data_type = 'enum'
                                    order by table_name,ordinal_position"""

    foreign_keys_query = """SELECT TABLE_NAME, COLUMN_NAME, REFERENCED_TABLE_NAME, REFERENCED_COLUMN_NAME
                                    FROM information_schema.KEY_COLUMN_USAGE
                                    WHERE TABLE_SCHEMA = %s AND REFERENCED_TABLE_NAME IS NOT NULL"""

    now_query = """SELECT NOW()"""

    @staticmethod
    def _parse_enum_values(column_type: str) -> list[str]:
        if not column_type or not column_type.lower().startswith("enum("):
            return []

        values: list[str] = []
        current: list[str] = []
        in_quote = False
        i = column_type.find("(") + 1

        while i < len(column_type):
            ch = column_type[i]

            if not in_quote:
                if ch == "'":
                    in_quote = True
                    current = []
                elif ch == ")":
                    break
            else:
                if ch == "\\" and i + 1 < len(column_type):
                    current.append(column_type[i + 1])
                    i += 1
                elif ch == "'":
                    if i + 1 < len(column_type) and column_type[i + 1] == "'":
                        current.append("'")
                        i += 1
                    else:
                        values.append("".join(current))
                        in_quote = False
                else:
                    current.append(ch)
            i += 1

        return values

    def __init__(
        self,
        database: str | None,
        user: str | None,
        password: str | None,
        host: str | None,
        port: int | None,
        socket: str | None,
        character_set: str | None,
        local_infile: bool | None,
        ssl: dict[str, Any] | None,
        ssh_user: str | None,
        ssh_host: str | None,
        ssh_port: int | None,
        ssh_password: str | None,
        ssh_key_filename: str | None,
        init_command: str | None = None,
        unbuffered: bool | None = None,
    ) -> None:
        self.dbname = database
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.socket = socket
        self.character_set = character_set
        self.local_infile = local_infile
        self.ssl = ssl
        self.server_info: ServerInfo | None = None
        self.connection_id: int | None = None
        self.ssh_user = ssh_user
        self.ssh_host = ssh_host
        self.ssh_port = ssh_port
        self.ssh_password = ssh_password
        self.ssh_key_filename = ssh_key_filename
        self.init_command = init_command
        self.unbuffered = unbuffered
        self.conn: Connection | None = None
        self.connect()

    def connect(
        self,
        database: str | None = None,
        user: str | None = None,
        password: str | None = None,
        host: str | None = None,
        port: int | None = None,
        socket: str | None = None,
        character_set: str | None = None,
        local_infile: bool | None = None,
        ssl: dict[str, Any] | None = None,
        ssh_host: str | None = None,
        ssh_port: int | None = None,
        ssh_user: str | None = None,
        ssh_password: str | None = None,
        ssh_key_filename: str | None = None,
        init_command: str | None = None,
        unbuffered: bool | None = None,
    ):
        db = database if database is not None else self.dbname
        user = user if user is not None else self.user
        password = password if password is not None else self.password
        host = host if host is not None else self.host
        port = port if port is not None else self.port
        socket = socket if socket is not None else self.socket
        character_set = character_set if character_set is not None else self.character_set
        local_infile = local_infile if local_infile is not None else self.local_infile
        ssl = ssl if ssl is not None else self.ssl
        ssh_user = ssh_user if ssh_user is not None else self.ssh_user
        ssh_host = ssh_host if ssh_host is not None else self.ssh_host
        ssh_port = ssh_port if ssh_port is not None else self.ssh_port
        ssh_password = ssh_password if ssh_password is not None else self.ssh_password
        ssh_key_filename = ssh_key_filename if ssh_key_filename is not None else self.ssh_key_filename
        init_command = init_command if init_command is not None else self.init_command
        unbuffered = unbuffered if unbuffered is not None else self.unbuffered
        _logger.debug(
            "Connection DB Params: \n"
            "\tdatabase: %r"
            "\tuser: %r"
            "\thost: %r"
            "\tport: %r"
            "\tsocket: %r"
            "\tcharacter_set: %r"
            "\tlocal_infile: %r"
            "\tssl: %r"
            "\tssh_user: %r"
            "\tssh_host: %r"
            "\tssh_port: %r"
            "\tssh_password: ***"
            "\tssh_key_filename: %r"
            "\tinit_command: %r"
            "\tunbuffered: %r",
            db,
            user,
            host,
            port,
            socket,
            character_set,
            local_infile,
            ssl,
            ssh_user,
            ssh_host,
            ssh_port,
            ssh_key_filename,
            init_command,
            unbuffered,
        )
        conv = conversions.copy()
        conv.update({
            FIELD_TYPE.TIMESTAMP: lambda obj: convert_datetime(obj) or obj,
            FIELD_TYPE.DATETIME: lambda obj: convert_datetime(obj) or obj,
            FIELD_TYPE.TIME: lambda obj: convert_time(obj) or obj,
            FIELD_TYPE.DATE: lambda obj: convert_date(obj) or obj,
        })

        defer_connect = False

        if ssh_host:
            defer_connect = True

        client_flag = pymysql.constants.CLIENT.INTERACTIVE
        if init_command and len(list(iocommands.split_queries(init_command))) > 1:
            client_flag |= pymysql.constants.CLIENT.MULTI_STATEMENTS
        client_flag |= pymysql.constants.CLIENT.HANDLE_EXPIRED_PASSWORDS

        ssl_context = None
        if ssl:
            ssl_context = self._create_ssl_ctx(ssl)

        connect_kwargs: dict[str, Any] = {
            "database": db,
            "user": user,
            "password": password or '',
            "host": host,
            "port": port or 0,
            "unix_socket": socket,
            "use_unicode": True,
            "charset": character_set or '',
            "autocommit": True,
            "client_flag": client_flag,
            "local_infile": local_infile or False,
            "conv": conv,
            "ssl": ssl_context,  # type: ignore[arg-type]
            "program_name": "mycli",
            "defer_connect": defer_connect,
            "init_command": init_command or None,
            "cursorclass": pymysql.cursors.SSCursor if unbuffered else pymysql.cursors.Cursor,
        }

        self.sandbox_mode = False
        try:
            conn = pymysql.connect(**connect_kwargs)  # type: ignore[misc]
        except pymysql.OperationalError as e:
            if e.args[0] == ER_MUST_CHANGE_PASSWORD:
                # Post-handshake queries (SET NAMES, SET AUTOCOMMIT, init_command)
                # fail with ER_MUST_CHANGE_PASSWORD in sandbox mode.
                # Reconnect with only the raw handshake.
                connect_kwargs['defer_connect'] = True
                connect_kwargs['autocommit'] = None
                connect_kwargs['init_command'] = None
                conn = pymysql.connect(**connect_kwargs)  # type: ignore[misc]
                self._connect_sandbox(conn)
                self.sandbox_mode = True
            else:
                raise

        if ssh_host and not self.sandbox_mode:
            ##### paramiko.Channel is a bad socket implementation overall if you want SSL through an SSH tunnel
            #####
            # instead let's open a tunnel and rewrite host:port to local bind
            try:
                chan = sshtunnel.SSHTunnelForwarder(
                    (ssh_host, ssh_port),
                    ssh_username=ssh_user,
                    ssh_pkey=ssh_key_filename,
                    ssh_password=ssh_password,
                    remote_bind_address=(host, port),
                )
                chan.start()

                conn.host = chan.local_bind_host
                conn.port = chan.local_bind_port
                conn.connect()
            except Exception as e:
                raise e

        if self.conn is not None:
            try:
                self.conn.close()
            except pymysql.err.Error:
                pass
        self.conn = conn
        # Update them after the connection is made to ensure that it was a
        # successful connection.
        self.dbname = db
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.socket = socket
        self.character_set = character_set
        self.ssl = ssl
        self.init_command = init_command
        self.unbuffered = unbuffered
        # retrieve connection id (skip in sandbox mode as queries will fail)
        if not self.sandbox_mode:
            self.reset_connection_id()
            server_info = ServerInfo.from_version_string(conn.server_version)  # type: ignore[attr-defined]
            if server_info.species == ServerSpecies.MySQL:
                if (doris_version := self._probe_doris_version()) is not None:
                    server_info = ServerInfo(ServerSpecies.Doris, doris_version)
            self.server_info = server_info

    def _probe_doris_version(self) -> str | None:
        """Query the server to check if it is Doris. Returns the Doris version string
        (e.g. '2.1.7') if confirmed, or None if the server is not Doris."""
        if self.sandbox_mode or self.conn is None:
            return None
        try:
            with self.conn.cursor() as cur:
                cur.execute("SELECT @@version_comment, @@version")
                row = cur.fetchone()
            if not row:
                return None
            version_comment = str(row[0] or "")
            version = str(row[1] or "")
            if "doris" not in version_comment.lower() and "doris" not in version.lower():
                return None
            # Prefer @@version_comment which usually carries the real Doris version string
            # (e.g. "doris-2.1.7-rc01"), then fall back to @@version.
            for candidate in (version_comment, version):
                ver_match = re.search(r"([0-9]+\.[0-9]+\.[0-9]+)", candidate)
                if ver_match:
                    return ver_match.group(1)
            # Doris confirmed but couldn't parse a version number
            return ""
        except Exception as e:
            _logger.debug("Doris detection failed: %s", e)
        return None

    def run(self, statement: str) -> Generator[SQLResult, None, None]:
        """Execute the sql in the database and return the results."""

        # Remove spaces and EOL
        statement = statement.strip()
        if not statement:  # Empty string
            yield SQLResult()

        # Split the sql into separate queries and run each one.
        # Unless it's saving a favorite query, in which case we
        # want to save them all together.
        if statement.startswith("\\fs"):
            components: Iterable[str] = [statement]
        else:
            components = iocommands.split_queries(statement)

        for sql in components:
            # \G is treated specially since we have to set the expanded output.
            if sql.endswith("\\G"):
                iocommands.set_expanded_output(True)
                sql = sql[:-2].strip()
            # \g is treated specially since we might want collapsed output when
            # auto vertical output is enabled
            elif sql.endswith('\\g'):
                iocommands.set_expanded_output(False)
                iocommands.set_forced_horizontal_output(True)
                sql = sql[:-2].strip()

            assert isinstance(self.conn, Connection)
            cur = self.conn.cursor()
            try:  # Special command
                _logger.debug("Trying a dbspecial command. sql: %r", sql)
                yield from execute(cur, sql)
            except CommandNotFound:  # Regular SQL
                _logger.debug("Regular sql statement. sql: %r", sql)
                cur.execute(sql)
                while True:
                    yield self.get_result(cur)

                    # PyMySQL returns an extra, empty result set with stored
                    # procedures. We skip it (rowcount is zero and no
                    # description).
                    if not cur.nextset() or (not cur.rowcount and cur.description is None):
                        break

    def get_result(self, cursor: Cursor) -> SQLResult:
        """Get the current result's data from the cursor."""
        preamble = header = None

        # cursor.description is not None for queries that return result sets,
        # e.g. SELECT or SHOW.
        plural = '' if cursor.rowcount == 1 else 's'
        if cursor.description:
            header = [x[0] for x in cursor.description]
            status = FormattedText([('', f'{cursor.rowcount} row{plural} in set')])
        else:
            _logger.debug("No rows in result.")
            status = FormattedText([('', f'Query OK, {cursor.rowcount} row{plural} affected')])

        if cursor.warning_count > 0:
            plural = '' if cursor.warning_count == 1 else 's'
            comma = FormattedText([('', ', ')])
            warning_count = FormattedText([('class:output.status.warning-count', f'{cursor.warning_count} warning{plural}')])
            status.extend(comma)
            status.extend(warning_count)

        return SQLResult(preamble=preamble, header=header, rows=cursor, status=status)

    def tables(self) -> Generator[tuple[str], None, None]:
        """Yields table names"""

        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Tables Query. sql: %r", self.tables_query)
            cur.execute(self.tables_query)
            yield from cur

    def table_columns(self, schema: str | None = None) -> Generator[tuple[str, str], None, None]:
        """Yields (table name, column name) pairs for *schema* (default: current database)."""
        target = schema if schema is not None else self.dbname
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Columns Query. sql: %r schema: %r", self.table_columns_query, target)
            cur.execute(self.table_columns_query, (target,))
            yield from cur

    def enum_values(self, schema: str | None = None) -> Generator[tuple[str, str, list[str]], None, None]:
        """Yields (table name, column name, enum values) tuples for *schema*."""
        target = schema if schema is not None else self.dbname
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Enum Values Query. sql: %r schema: %r", self.enum_values_query, target)
            cur.execute(self.enum_values_query, (target,))
            for table_name, column_name, column_type in cur:
                values = self._parse_enum_values(column_type)
                if values:
                    yield (table_name, column_name, values)

    def foreign_keys(self, schema: str | None = None) -> Generator[tuple[str, str, str, str], None, None]:
        """Yields (table_name, column_name, referenced_table_name, referenced_column_name) tuples for *schema*."""
        target = schema if schema is not None else self.dbname
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Foreign Keys Query. sql: %r schema: %r", self.foreign_keys_query, target)
            try:
                cur.execute(self.foreign_keys_query, (target,))
                yield from cur
            except Exception as e:
                _logger.error('No foreign key completions due to %r', e)

    def databases(self) -> list[str]:
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Databases Query. sql: %r", self.databases_query)
            cur.execute(self.databases_query)
            return [x[0] for x in cur.fetchall()]

    def functions(self, schema: str | None = None) -> Generator[tuple[str, str], None, None]:
        """Yields tuples of (schema_name, function_name) for *schema*."""

        target = schema if schema is not None else self.dbname
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Functions Query. sql: %r schema: %r", self.functions_query, target)
            cur.execute(self.functions_query, (target,))
            yield from cur

    def procedures(self, schema: str | None = None) -> Generator[tuple, None, None]:
        """Yields tuples of (procedure_name, ) for *schema*."""

        target = schema if schema is not None else self.dbname
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Procedures Query. sql: %r schema: %r", self.procedures_query, target)
            try:
                cur.execute(self.procedures_query, (target,))
            except pymysql.DatabaseError as e:
                _logger.error('No procedure completions due to %r', e)
                yield ()
            else:
                yield from cur

    def character_sets(self) -> Generator[tuple, None, None]:
        """Yields tuples of (character_set_name, )"""

        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Character sets Query. sql: %r", self.character_sets_query)
            try:
                cur.execute(self.character_sets_query)
            except pymysql.DatabaseError as e:
                _logger.error('No character_set completions due to %r', e)
                yield ()
            else:
                yield from cur

    def collations(self) -> Generator[tuple, None, None]:
        """Yields tuples of (collation_name, )"""

        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Collations Query. sql: %r", self.collations_query)
            try:
                cur.execute(self.collations_query)
            except pymysql.DatabaseError as e:
                _logger.error('No collations completions due to %r', e)
                yield ()
            else:
                yield from cur

    def show_candidates(self) -> Generator[tuple, None, None]:
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Show Query. sql: %r", self.show_candidates_query)
            try:
                cur.execute(self.show_candidates_query)
            except pymysql.DatabaseError as e:
                _logger.error("No show completions due to %r", e)
                yield ()
            else:
                for row in cur:
                    yield (row[0].split(None, 1)[-1],)

    def users(self) -> Generator[tuple, None, None]:
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Users Query. sql: %r", self.users_query)
            try:
                cur.execute(self.users_query)
            except pymysql.DatabaseError as e:
                _logger.error("No user completions due to %r", e)
                yield ()
            else:
                yield from cur

    def now(self) -> datetime.datetime:
        assert isinstance(self.conn, Connection)
        with self.conn.cursor() as cur:
            _logger.debug("Now Query. sql: %r", self.now_query)
            cur.execute(self.now_query)
            if one := cur.fetchone():
                return one[0]
            else:
                return datetime.datetime.now()

    def get_connection_id(self) -> int | None:
        if not self.connection_id:
            self.reset_connection_id()
        return self.connection_id

    def reset_connection_id(self) -> None:
        # Remember current connection id
        _logger.debug("Get current connection id")
        try:
            results = self.run("select connection_id()")
            for result in results:
                cur = result.rows
                if isinstance(cur, Cursor):
                    v = cur.fetchone()
                    self.connection_id = v[0] if v is not None else -1
                else:
                    raise ValueError
        except Exception as e:
            # See #1054
            self.connection_id = -1
            _logger.error("Failed to get connection id: %s", e)
        else:
            _logger.debug("Current connection id: %s", self.connection_id)

    def change_db(self, db: str) -> None:
        assert isinstance(self.conn, Connection)
        self.conn.select_db(db)
        self.dbname = db

    @staticmethod
    def _connect_sandbox(conn: Connection) -> None:
        """Connect in sandbox mode, performing only the handshake.

        pymysql's normal connect() runs post-handshake queries (SET NAMES,
        SET AUTOCOMMIT, init_command) that all fail with ER_MUST_CHANGE_PASSWORD
        in sandbox mode.  This method performs the raw socket connection and
        authentication handshake only.
        """
        # Reuse pymysql internals for the handshake + auth, but
        # temporarily stub out set_character_set so it becomes a no-op.
        original_set_charset = conn.set_character_set
        conn.set_character_set = lambda *_args, **_kwargs: None  # type: ignore[assignment]
        try:
            conn.connect()
        finally:
            conn.set_character_set = original_set_charset  # type: ignore[assignment]

    def _create_ssl_ctx(self, sslp: dict) -> ssl.SSLContext:
        ca = sslp.get("ca")
        capath = sslp.get("capath")
        hasnoca = ca is None and capath is None
        ctx = ssl.create_default_context(cafile=ca, capath=capath)
        ctx.check_hostname = not hasnoca and sslp.get("check_hostname", True)
        ctx.verify_mode = ssl.CERT_NONE if hasnoca else ssl.CERT_REQUIRED
        if "cert" in sslp:
            ctx.load_cert_chain(sslp["cert"], keyfile=sslp.get("key"))
        if "cipher" in sslp:
            ctx.set_ciphers(sslp["cipher"])

        ctx.minimum_version = ssl.TLSVersion.TLSv1_2

        if "tls_version" in sslp:
            tls_version = sslp["tls_version"]

            if tls_version == "TLSv1":
                ctx.minimum_version = ssl.TLSVersion.TLSv1
                ctx.maximum_version = ssl.TLSVersion.TLSv1
            elif tls_version == "TLSv1.1":
                ctx.minimum_version = ssl.TLSVersion.TLSv1_1
                ctx.maximum_version = ssl.TLSVersion.TLSv1_1
            elif tls_version == "TLSv1.2":
                ctx.minimum_version = ssl.TLSVersion.TLSv1_2
                ctx.maximum_version = ssl.TLSVersion.TLSv1_2
            elif tls_version == "TLSv1.3":
                ctx.minimum_version = ssl.TLSVersion.TLSv1_3
                ctx.maximum_version = ssl.TLSVersion.TLSv1_3
            else:
                _logger.error("Invalid tls version: %s", tls_version)

        return ctx

    def close(self) -> None:
        if self.conn is not None:
            try:
                self.conn.close()
            except pymysql.err.Error:
                pass



from pygments.token import Keyword, Name

from mycli.lexer import MyCliLexer


def test_mysql_lexer_keeps_identifiers_starting_with_set_together():
    tokens = list(MyCliLexer().get_tokens("SELECT * FROM settings_123 WHERE id = 123;"))

    assert (Name, "settings_123") in tokens
    assert (Keyword, "set") not in tokens



from __future__ import annotations

import locale
import logging
import os
import re
import shlex
import subprocess
from time import sleep
from typing import Any, Generator

import click
from configobj import ConfigObj
from prompt_toolkit.formatted_text import ANSI, FormattedText, to_plain_text
from pymysql.cursors import Cursor
import pyperclip
import sqlparse

from mycli.compat import WIN
from mycli.packages.interactive_utils import confirm_destructive_query
from mycli.packages.special.delimitercommand import DelimiterCommand
from mycli.packages.special.favoritequeries import FavoriteQueries
from mycli.packages.special.main import COMMANDS as SPECIAL_COMMANDS
from mycli.packages.special.main import ArgType, SpecialCommandAlias, special_command
from mycli.packages.special.main import execute as special_execute
from mycli.packages.special.utils import handle_cd_command
from mycli.packages.sqlresult import SQLResult

sqlparse.engine.grouping.MAX_GROUPING_DEPTH = None  # type: ignore[assignment]
sqlparse.engine.grouping.MAX_GROUPING_TOKENS = None  # type: ignore[assignment]

TIMING_ENABLED = False
use_expanded_output = False
force_horizontal_output = False
PAGER_ENABLED = True
SHOW_FAVORITE_QUERY = True
tee_file = None
once_file = None
written_to_once_file = False
PIPE_ONCE: dict[str, Any] = {
    'process': None,
    'stdin': [],
    'stdout_file': None,
    'stdout_mode': None,
}
delimiter_command = DelimiterCommand()
favoritequeries = FavoriteQueries(ConfigObj())
DESTRUCTIVE_KEYWORDS: list[str] = []
SHOW_WARNINGS_ENABLED: bool = False


def set_favorite_queries(config):
    global favoritequeries
    favoritequeries = FavoriteQueries(config)


def set_timing_enabled(val: bool) -> None:
    global TIMING_ENABLED
    TIMING_ENABLED = val


def set_pager_enabled(val: bool) -> None:
    global PAGER_ENABLED
    PAGER_ENABLED = val


def is_pager_enabled() -> bool:
    return PAGER_ENABLED


def set_show_favorite_query(val: bool) -> None:
    global SHOW_FAVORITE_QUERY
    SHOW_FAVORITE_QUERY = val


def is_show_favorite_query() -> bool:
    return SHOW_FAVORITE_QUERY


def set_destructive_keywords(val: list[str]) -> None:
    global DESTRUCTIVE_KEYWORDS
    DESTRUCTIVE_KEYWORDS = val


def set_show_warnings_enabled(val: bool) -> None:
    global SHOW_WARNINGS_ENABLED
    SHOW_WARNINGS_ENABLED = val


def is_show_warnings_enabled() -> bool:
    return SHOW_WARNINGS_ENABLED


@special_command(
    'warnings',
    'warnings',
    'Enable automatic warnings display.',
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias('\\W', case_sensitive=True)],
)
def enable_show_warnings() -> Generator[SQLResult, None, None]:
    global SHOW_WARNINGS_ENABLED
    SHOW_WARNINGS_ENABLED = True
    msg = "Show warnings enabled."
    yield SQLResult(status=msg)


@special_command(
    'nowarnings',
    'nowarnings',
    'Disable automatic warnings display.',
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias('\\w', case_sensitive=True)],
)
def disable_show_warnings() -> Generator[SQLResult, None, None]:
    global SHOW_WARNINGS_ENABLED
    SHOW_WARNINGS_ENABLED = False
    msg = 'Show warnings disabled.'
    yield SQLResult(status=msg)


@special_command(
    "pager",
    "pager [command]",
    "Set pager to [command]. Print query results via pager.",
    arg_type=ArgType.PARSED_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias("\\P", case_sensitive=True)],
)
def set_pager(arg: str, **_) -> list[SQLResult]:
    if arg:
        os.environ["PAGER"] = arg
        msg = f"PAGER set to {arg}."
        set_pager_enabled(True)
    else:
        if "PAGER" in os.environ:
            msg = f"PAGER set to {os.environ['PAGER']}."
        else:
            # This uses click's default per echo_via_pager.
            msg = "Pager enabled."
        set_pager_enabled(True)

    return [SQLResult(status=msg)]


@special_command(
    "nopager",
    "nopager",
    "Disable pager; print to stdout.",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias("\\n", case_sensitive=True)],
)
def disable_pager() -> list[SQLResult]:
    set_pager_enabled(False)
    return [SQLResult(status="Pager disabled.")]


@special_command(
    "\\timing",
    "\\timing",
    "Toggle timing of queries.",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias("\\t", case_sensitive=True)],
)
def toggle_timing() -> list[SQLResult]:
    global TIMING_ENABLED
    TIMING_ENABLED = not TIMING_ENABLED
    message = "Timing is "
    message += "on." if TIMING_ENABLED else "off."
    return [SQLResult(status=message)]


def is_timing_enabled() -> bool:
    return TIMING_ENABLED


def set_expanded_output(val: bool) -> None:
    global use_expanded_output
    use_expanded_output = val


def is_expanded_output() -> bool:
    return use_expanded_output


def set_forced_horizontal_output(val: bool) -> None:
    global force_horizontal_output
    force_horizontal_output = val


def forced_horizontal() -> bool:
    return force_horizontal_output


_logger = logging.getLogger(__name__)


def editor_command(command: str) -> bool:
    """
    Is this an external editor command?
    :param command: string
    """
    # special case: allow help on the \edit command
    if re.match(r'^([Hh][Ee][Ll][Pp])\s+(\\e|\\edit)\s*(;|\\G|\\g)?\s*$', command):
        return False
    # It is possible to have `\e filename` or `SELECT * FROM \e`. So we check
    # for both conditions.
    return (
        command.strip().endswith("\\e")
        or command.strip().startswith("\\e ")
        or command.strip().endswith("\\edit")
        or command.strip().startswith("\\edit ")
    )


def get_filename(sql: str) -> str | None:
    if sql.strip().startswith("\\e ") or sql.strip().startswith("\\edit "):
        command, _, filename = sql.partition(" ")
        return filename.strip() or None
    else:
        return None


def get_editor_query(sql: str) -> str:
    """Get the query part of an editor command."""
    sql = sql.strip()

    # The reason we can't simply do .strip('\e') is that it strips characters,
    # not a substring. So it'll strip "e" in the end of the sql also!
    # Ex: "select * from style\e" -> "select * from styl".
    pattern = re.compile(r"(\\e$|\\edit$)")
    while pattern.search(sql):
        sql = pattern.sub("", sql)

    return sql


def open_external_editor(filename: str | None = None, sql: str | None = None) -> tuple[str, str | None]:
    """Open external editor, wait for the user to type in their query, return
    the query.
    """

    filename = filename.strip().split(" ", 1)[0] if filename else None
    sql = sql or ""
    MARKER = "# Type your query above this line.\n"

    if filename:
        query = ''
        message = None
        click.edit(filename=filename)
        try:
            with open(filename, 'r') as f:
                query = f.read()
        except IOError:
            message = f'Error reading file: {filename}'
        return (query.rstrip('\n'), message)

    # Populate the editor buffer with the partial sql (if available) and a
    # placeholder comment.
    query = click.edit(f"{sql}\n\n{MARKER}", extension=".sql") or ''

    if query:
        query = query.split(MARKER, 1)[0].rstrip("\n")
    else:
        # Don't return None for the caller to deal with.
        # Empty string is ok.
        query = sql

    return (query, None)


def clip_command(command: str) -> bool:
    """Is this a clip command?

    :param command: string

    """
    # It is possible to have `\clip` or `SELECT * FROM \clip`. So we check
    # for both conditions.
    return command.strip().endswith("\\clip") or command.strip().startswith("\\clip")


def get_clip_query(sql: str) -> str:
    """Get the query part of a clip command."""
    sql = sql.strip()

    # The reason we can't simply do .strip('\clip') is that it strips characters,
    # not a substring. So it'll strip "c" in the end of the sql also!
    pattern = re.compile(r"(^\\clip|\\clip$)")
    while pattern.search(sql):
        sql = pattern.sub("", sql)

    return sql


def copy_query_to_clipboard(sql: str | None = None) -> str | None:
    """Send query to the clipboard."""

    sql = sql or ""
    message = None

    try:
        pyperclip.copy(f"{sql}")
    except RuntimeError as e:
        message = f"Error clipping query: {e}."

    return message


def set_redirect(command_part: str | None, file_operator_part: str | None, file_part: str | None) -> list[tuple]:
    if command_part:
        if file_part:
            PIPE_ONCE['stdout_file'] = file_part
            PIPE_ONCE['stdout_mode'] = 'w' if file_operator_part == '>' else 'a'
        return set_pipe_once(command_part)
    elif file_operator_part == '>':
        return set_once(f'-o {file_part}')
    else:
        return set_once(file_part)


@special_command(
    "\\f",
    "\\f [name [args..]]",
    "List or execute favorite queries.",
    arg_type=ArgType.PARSED_QUERY,
    case_sensitive=True,
)
def execute_favorite_query(cur: Cursor, arg: str, **_) -> Generator[SQLResult, None, None]:
    if arg == "":
        yield from list_favorite_queries()
        return

    # Parse out favorite name and optional substitution parameters
    name, _separator, arg_str = arg.partition(" ")
    args = shlex.split(arg_str)

    query = FavoriteQueries.instance.get(name)
    if query is None:
        message = f"No favorite query: {name}"
        yield SQLResult(status=message)
    else:
        query, arg_error = subst_favorite_query_args(query, args)
        if query is None:
            yield SQLResult(status=arg_error)
        else:
            for sql in sqlparse.split(query):
                sql = sql.rstrip(";")
                preamble = f"> {sql}" if is_show_favorite_query() else None
                is_special = False
                for special in SPECIAL_COMMANDS:
                    if sql.lower().startswith(special.lower()):
                        is_special = True
                        break
                if is_special:
                    for result in special_execute(cur, sql):
                        result.preamble = preamble
                        # special_execute() already returns a SQLResult
                        yield result
                else:
                    cur.execute(sql)
                    if cur.description:
                        header = [x[0] for x in cur.description]
                        yield SQLResult(preamble=preamble, header=header, rows=cur)
                    else:
                        yield SQLResult(preamble=preamble)


def list_favorite_queries() -> list[SQLResult]:
    """List of all favorite queries."""

    header = ["Name", "Query"]
    rows = [(r, FavoriteQueries.instance.get(r)) for r in FavoriteQueries.instance.list()]

    if not rows:
        status = "\nNo favorite queries found." + FavoriteQueries.instance.usage
    else:
        status = ""
    return [SQLResult(header=header, rows=rows, status=status)]


def subst_favorite_query_args(query: str, args: list[str]) -> list[str | None]:
    """replace positional parameters ($1...$N) in query."""
    for idx, val in enumerate(args):
        subst_var = "$" + str(idx + 1)
        if subst_var not in query:
            return [None, "query does not have substitution parameter " + subst_var + ":\n  " + query]

        query = query.replace(subst_var, val)

    match = re.search(r"\$\d+", query)
    if match:
        return [None, "missing substitution for " + match.group(0) + " in query:\n  " + query]

    return [query, None]


@special_command(
    "\\fs",
    "\\fs <name> <query>",
    "Save a favorite query.",
)
def save_favorite_query(arg: str, **_) -> list[SQLResult]:
    """Save a new favorite query."""

    usage = "Syntax: \\fs name query.\n\n" + FavoriteQueries.instance.usage
    if not arg:
        return [SQLResult(status=usage)]

    name, _separator, query = arg.partition(" ")

    # If either name or query is missing then print the usage and complain.
    if (not name) or (not query):
        return [SQLResult(status=f"{usage} Err: Both name and query are required.")]

    FavoriteQueries.instance.save(name, query)
    return [SQLResult(status="Saved.")]


@special_command(
    "\\fd",
    "\\fd <name>",
    "Delete a favorite query.",
)
def delete_favorite_query(arg: str, **_) -> list[SQLResult]:
    """Delete an existing favorite query."""
    usage = "Syntax: \\fd name.\n\n" + FavoriteQueries.instance.usage
    if not arg:
        return [SQLResult(status=usage)]

    status = FavoriteQueries.instance.delete(arg)

    return [SQLResult(status=status)]


@special_command(
    "system",
    "system [-r] <command>",
    "Execute a system shell command (raw mode with -r).",
)
def execute_system_command(arg: str, **_) -> list[SQLResult]:
    """Execute a system shell command."""
    usage = "Syntax: system [-r] [command].\n-r denotes \"raw\" mode, in which output is passed through without formatting."

    IMPLICIT_RAW_MODE_COMMANDS = {
        'clear',
        'vim',
        'vi',
        'bash',
        'zsh',
    }

    if not arg.strip():
        return [SQLResult(status=usage)]

    try:
        command = shlex.split(arg.strip(), posix=not WIN)
    except ValueError as e:
        return [SQLResult(status=f"Cannot parse system command: {e}")]

    raw = False
    if command[0] == '-r':
        command.pop(0)
        raw = True
    elif command[0].lower() in IMPLICIT_RAW_MODE_COMMANDS:
        raw = True

    if not command:
        return [SQLResult(status=usage)]

    if command[0].lower() == 'cd':
        ok, error_message = handle_cd_command(command)
        if not ok:
            return [SQLResult(status=error_message)]
        return [SQLResult()]

    try:
        if raw:
            completed_process = subprocess.run(command, check=False)
            if completed_process.returncode:
                return [SQLResult(status=f'Command exited with return code {completed_process.returncode}')]
            else:
                return [SQLResult()]
        else:
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            try:
                output, error = process.communicate(timeout=60)
            except subprocess.TimeoutExpired:
                process.kill()
                output, error = process.communicate()
            response = output if not error else error
            encoding = locale.getpreferredencoding(False)
            response_str = response.decode(encoding)
            if process.returncode:
                status = f'Command exited with return code {process.returncode}'
            else:
                status = None
            return [SQLResult(preamble=response_str, status=status)]
    except OSError as e:
        return [SQLResult(status=f"OSError: {e.strerror}")]


def parseargfile(arg: str) -> tuple[str, str]:
    if arg.startswith("-o "):
        mode = "w"
        filename = arg[3:]
    else:
        mode = "a"
        filename = arg

    if not filename:
        raise TypeError("You must provide a filename.")

    return (os.path.expanduser(filename), mode)


@special_command(
    "tee",
    "tee [-o] <filename>",
    "Append all results to an output file (overwrite using -o).",
)
def set_tee(arg: str, **_) -> list[SQLResult]:
    global tee_file

    try:
        tee_file = open(*parseargfile(arg))
    except (IOError, OSError) as e:
        raise OSError(f"Cannot write to file '{e.filename}': {e.strerror}") from e

    return [SQLResult(status="")]


def close_tee() -> None:
    global tee_file
    if tee_file:
        tee_file.close()
        tee_file = None


@special_command(
    "notee",
    "notee",
    "Stop writing results to an output file.",
)
def no_tee(arg: str, **_) -> list[SQLResult]:
    close_tee()
    return [SQLResult(status="")]


def write_tee(output: str | ANSI | FormattedText, nl: bool = True) -> None:
    global tee_file
    if not tee_file:
        return
    click.echo(to_plain_text(output), file=tee_file, nl=False)
    if nl:
        click.echo('\n', file=tee_file, nl=False)
    tee_file.flush()


@special_command(
    "\\once",
    "\\once [-o] <filename>",
    "Append next result to an output file (overwrite using -o).",
    aliases=[SpecialCommandAlias("\\o", case_sensitive=False)],
)
def set_once(arg: str, **_) -> list[SQLResult]:
    global once_file, written_to_once_file

    try:
        once_file = open(*parseargfile(arg))
    except (IOError, OSError) as e:
        raise OSError(f"Cannot write to file '{e.filename}': {e.strerror}") from e
    written_to_once_file = False

    return [SQLResult(status="")]


def is_redirected() -> bool:
    return bool(once_file or PIPE_ONCE['process'])


def write_once(output: str) -> None:
    global once_file, written_to_once_file
    if output and once_file:
        click.echo(output, file=once_file, nl=False)
        click.echo("\n", file=once_file, nl=False)
        once_file.flush()
        written_to_once_file = True


def unset_once_if_written(post_redirect_command: str) -> None:
    """Unset the once file, if it has been written to."""
    global once_file, written_to_once_file
    if written_to_once_file and once_file:
        once_filename = once_file.name
        once_file.close()
        once_file = None
        _run_post_redirect_hook(post_redirect_command, once_filename)


def _run_post_redirect_hook(post_redirect_command: str, filename: str) -> None:
    if not post_redirect_command:
        return
    post_cmd = post_redirect_command.format(shlex.quote(filename))
    try:
        subprocess.run(
            post_cmd,
            shell=True,
            check=True,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except Exception as e:
        raise OSError(f"Redirect post hook failed: {e}") from e


@special_command(
    "\\pipe_once",
    "\\pipe_once <command>",
    "Send next result to a subprocess.",
    aliases=[SpecialCommandAlias("\\|", case_sensitive=False)],
)
def set_pipe_once(arg: str, **_) -> list[SQLResult]:
    if not arg:
        raise OSError("pipe_once requires a command")
    if WIN:
        # best effort, no chaining
        pipe_once_cmd = shlex.split(arg)
    else:
        # to support chaining
        pipe_once_cmd = ['sh', '-c', arg]
    PIPE_ONCE['stdin'] = []
    PIPE_ONCE['process'] = subprocess.Popen(
        pipe_once_cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="UTF-8",
        universal_newlines=True,
    )
    return [SQLResult(status="")]


def write_pipe_once(line: str) -> None:
    if line and PIPE_ONCE['process']:
        PIPE_ONCE['stdin'].append(line)


def flush_pipe_once_if_written(post_redirect_command: str) -> None:
    """Flush the pipe_once cmd, if lines have been written."""
    if not PIPE_ONCE['process']:
        return
    if not PIPE_ONCE['stdin']:
        return
    try:
        (stdout_data, stderr_data) = PIPE_ONCE['process'].communicate(input='\n'.join(PIPE_ONCE['stdin']) + '\n', timeout=60)
    except subprocess.TimeoutExpired:
        PIPE_ONCE['process'].kill()
        (stdout_data, stderr_data) = PIPE_ONCE['process'].communicate()
    if stdout_data:
        if PIPE_ONCE['stdout_file']:
            with open(PIPE_ONCE['stdout_file'], PIPE_ONCE['stdout_mode']) as f:
                print(stdout_data, file=f)
            _run_post_redirect_hook(post_redirect_command, PIPE_ONCE['stdout_file'])
        else:
            click.secho(stdout_data.rstrip('\n'))
    if stderr_data:
        click.secho(stderr_data.rstrip('\n'), err=True, fg='red')
    if returncode := PIPE_ONCE['process'].returncode:
        PIPE_ONCE['process'] = None
        PIPE_ONCE['stdin'] = []
        PIPE_ONCE['stdout_file'] = None
        PIPE_ONCE['stdout_mode'] = None
        raise OSError(f'process exited with nonzero code {returncode}')
    PIPE_ONCE['process'] = None
    PIPE_ONCE['stdin'] = []
    PIPE_ONCE['stdout_file'] = None
    PIPE_ONCE['stdout_mode'] = None


@special_command(
    "watch",
    "watch [seconds] [-c] <query>",
    "Execute query every [seconds] seconds (5 by default).",
)
def watch_query(arg: str, **kwargs) -> Generator[SQLResult, None, None]:
    usage = """Syntax: watch [seconds] [-c] query.
    * seconds: The interval at the query will be repeated, in seconds.
               By default 5.
    * -c: Clears the screen between every iteration.
"""
    if not arg:
        yield SQLResult(status=usage)
        return
    seconds = 5.0
    clear_screen = False
    statement = None
    while statement is None:
        arg = arg.strip()
        if not arg:
            # Oops, we parsed all the arguments without finding a statement
            yield SQLResult(status=usage)
            return
        (left_arg, _, right_arg) = arg.partition(" ")
        arg = right_arg
        try:
            seconds = float(left_arg)
            continue
        except ValueError:
            pass
        if left_arg == "-c":
            clear_screen = True
            continue
        statement = f"{left_arg} {arg}"
    destructive_prompt = confirm_destructive_query(DESTRUCTIVE_KEYWORDS, statement)
    if destructive_prompt is False:
        click.secho("Wise choice!")
        return
    elif destructive_prompt is True:
        click.secho("Your call!")
    cur = kwargs["cur"]
    sql_list = [(sql.rstrip(";"), f"> {sql}") for sql in sqlparse.split(statement)]
    old_pager_enabled = is_pager_enabled()
    while True:
        if clear_screen:
            click.clear()
        try:
            # Somewhere in the code the pager its activated after every yield,
            # so we disable it in every iteration
            set_pager_enabled(False)
            for sql, preamble in sql_list:
                cur.execute(sql)
                command: dict[str, str | float] = {
                    "name": "watch",
                    "seconds": seconds,
                }
                if cur.description:
                    header = [x[0] for x in cur.description]
                    yield SQLResult(preamble=preamble, header=header, rows=cur, command=command)
                else:
                    yield SQLResult(preamble=preamble, command=command)
            sleep(seconds)
        except KeyboardInterrupt:
            # This prints the Ctrl-C character in its own line, which prevents
            # to print a line with the cursor positioned behind the prompt
            click.secho("", nl=True)
            return
        finally:
            set_pager_enabled(old_pager_enabled)


@special_command(
    "delimiter",
    "delimiter <string>",
    "Change end-of-statement delimiter.",
)
def set_delimiter(arg: str, **_) -> list[SQLResult]:
    return delimiter_command.set(arg)


def get_current_delimiter() -> str:
    return delimiter_command.current


def split_queries(input_str: str) -> Generator[str, None, None]:
    yield from delimiter_command.queries_iter(input_str)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import functools
from functools import partial
import html
from importlib import resources
import os
import random
import re
import subprocess
import sys
import time
import traceback
from typing import TYPE_CHECKING, Any, Generator
from xml.parsers.expat import ExpatError

import click
import prompt_toolkit
from prompt_toolkit.application.current import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory, ThreadedAutoSuggest
from prompt_toolkit.completion import DynamicCompleter
from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
from prompt_toolkit.filters import Condition, has_focus, is_done
from prompt_toolkit.formatted_text import (
    ANSI,
    HTML,
    FormattedText,
    to_formatted_text,
    to_plain_text,
)
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.processors import ConditionalProcessor, HighlightMatchingBracketProcessor
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.shortcuts import CompleteStyle, PromptSession
import pymysql
from pymysql.cursors import Cursor

import mycli as mycli_package
from mycli.clibuffer import cli_is_multiline
from mycli.clistyle import style_factory_ptoolkit
from mycli.clitoolbar import create_toolbar_tokens_func
from mycli.constants import (
    DEFAULT_HOST,
    DEFAULT_WIDTH,
    ER_MUST_CHANGE_PASSWORD,
    HOME_URL,
    ISSUES_URL,
)
from mycli.key_bindings import mycli_bindings
from mycli.lexer import MyCliLexer
from mycli.packages import special
from mycli.packages.filepaths import dir_path_exists
from mycli.packages.hybrid_redirection import get_redirect_components, is_redirect_command
from mycli.packages.interactive_utils import confirm, confirm_destructive_query
from mycli.packages.key_binding_utils import (
    handle_clip_command,
    handle_editor_command,
)
from mycli.packages.ptoolkit.history import FileHistoryWithTimestamp
from mycli.packages.special.utils import format_uptime, get_ssl_version, get_uptime, get_warning_count
from mycli.packages.sql_utils import (
    extract_new_password,
    is_dropping_database,
    is_mutating,
    is_password_change,
    is_sandbox_allowed,
    is_select,
    need_completion_refresh,
    need_completion_reset,
)
from mycli.packages.sqlresult import SQLResult
from mycli.packages.string_utils import sanitize_terminal_title
from mycli.sqlexecute import SQLExecute
from mycli.types import Query

if TYPE_CHECKING:
    from prompt_toolkit.formatted_text import AnyFormattedText

    from mycli.client import MyCli


SUPPORT_INFO = f"Home: {HOME_URL}\nBug tracker: {ISSUES_URL}"
MIN_COMPLETION_TRIGGER = 1
_PROMPT_TARGETS: dict[int, 'MyCli'] = {}


@dataclass(slots=True)
class ReplState:
    iterations: int = 0
    mutating: bool = False


@Condition
def complete_while_typing_filter() -> bool:
    """Whether enough characters have been typed to trigger completion.

    Written in a verbose way, with a string slice, for efficiency."""
    if MIN_COMPLETION_TRIGGER <= 1:
        return True
    app = get_app()
    text = app.current_buffer.text.lstrip()
    text_len = len(text)
    if text_len < MIN_COMPLETION_TRIGGER:
        return False
    last_word = text[-MIN_COMPLETION_TRIGGER:]
    if len(last_word) == text_len:
        return text_len >= MIN_COMPLETION_TRIGGER
    if text[:6].lower() in ['source', r'\.']:
        # Different word characters for paths; see comment below.
        # In fact, it might be nice if paths had a different threshold.
        return not bool(re.search(r'[\s!-,:-@\[-^\{\}-]', last_word))
    else:
        # This is "whitespace and all punctuation except underscore and backtick"
        # acting as word breaks, but it would be neat if we could complete differently
        # when inside a backtick, accepting all legal characters towards the trigger
        # limit. We would have to parse the statement, or at least go back more
        # characters, costing performance. This still works within a backtick! So
        # long as there are three trailing non-punctuation characters.
        return not bool(re.search(r'[\s!-/:-@\[-^\{-~]', last_word))


def _create_history(mycli: 'MyCli') -> FileHistoryWithTimestamp | None:
    history_file = os.path.expanduser(os.environ.get('MYCLI_HISTFILE', mycli.config['main'].get('history_file', '~/.mycli-history')))
    if dir_path_exists(history_file):
        return FileHistoryWithTimestamp(history_file)

    mycli.echo(
        f'Error: Unable to open the history file "{history_file}". Your query history will not be saved.',
        err=True,
        fg='red',
    )
    return None


def _show_startup_banner(
    mycli: 'MyCli',
    sqlexecute: SQLExecute,
) -> None:
    if mycli.verbosity < 0:
        return

    if sqlexecute.server_info is not None:
        print(sqlexecute.server_info)
    print('mycli', mycli_package.__version__)
    print(SUPPORT_INFO)
    if random.random() <= 0.25:
        print('Thanks to the sponsor —', _sponsors_picker())
    elif random.random() <= 0.5:
        print('Thanks to the contributor —', _contributors_picker())
    else:
        print('Tip —', _tips_picker())


def set_all_external_titles(mycli: 'MyCli') -> None:
    set_external_terminal_tab_title(mycli)
    set_external_terminal_window_title(mycli)
    set_external_multiplex_window_title(mycli)
    set_external_multiplex_pane_title(mycli)


def set_external_terminal_tab_title(mycli: 'MyCli') -> None:
    if not mycli.terminal_tab_title_format:
        return
    if not mycli.prompt_session:
        return
    if not sys.stderr.isatty():
        return
    title = sanitize_terminal_title(
        render_prompt_string(
            mycli,
            mycli.terminal_tab_title_format,
            mycli.prompt_session.app.render_counter,
        )
    )
    print(f'\x1b]1;{title}\a', file=sys.stderr, end='')
    sys.stderr.flush()


def set_external_terminal_window_title(mycli: 'MyCli') -> None:
    if not mycli.terminal_window_title_format:
        return
    if not mycli.prompt_session:
        return
    if not sys.stderr.isatty():
        return
    title = sanitize_terminal_title(
        render_prompt_string(
            mycli,
            mycli.terminal_window_title_format,
            mycli.prompt_session.app.render_counter,
        )
    )
    print(f'\x1b]2;{title}\a', file=sys.stderr, end='')
    sys.stderr.flush()


def set_external_multiplex_window_title(mycli: 'MyCli') -> None:
    if not mycli.multiplex_window_title_format:
        return
    if not os.getenv('TMUX'):
        return
    if not mycli.prompt_session:
        return
    title = sanitize_terminal_title(
        render_prompt_string(
            mycli,
            mycli.multiplex_window_title_format,
            mycli.prompt_session.app.render_counter,
        )
    )
    try:
        subprocess.run(
            ['tmux', 'rename-window', title],
            check=False,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        pass


def set_external_multiplex_pane_title(mycli: 'MyCli') -> None:
    if not mycli.multiplex_pane_title_format:
        return
    if not os.getenv('TMUX'):
        return
    if not mycli.prompt_session:
        return
    if not sys.stderr.isatty():
        return
    title = sanitize_terminal_title(
        render_prompt_string(
            mycli,
            mycli.multiplex_pane_title_format,
            mycli.prompt_session.app.render_counter,
        )
    )
    print(f'\x1b]2;{title}\x1b\\', file=sys.stderr, end='')
    sys.stderr.flush()


def get_custom_toolbar(
    mycli: 'MyCli',
    toolbar_format: str,
) -> FormattedText:
    if not mycli.prompt_session:
        return to_formatted_text('')
    if not mycli.prompt_session.app:
        return to_formatted_text('')
    if mycli.prompt_session.app.current_buffer.text:
        return mycli.last_custom_toolbar_message
    mycli.last_custom_toolbar_message = render_prompt_string(
        mycli,
        toolbar_format,
        mycli.prompt_session.app.render_counter,
    )
    return mycli.last_custom_toolbar_message


def maybe_html_escape(string: str, is_html: bool) -> str:
    if is_html:
        return html.escape(string, quote=False)
    return string


@functools.lru_cache(maxsize=256)
def render_prompt_string(
    mycli: 'MyCli',
    string: str,
    _render_counter: int,
) -> FormattedText:
    sqlexecute = mycli.sqlexecute
    assert sqlexecute is not None
    if mycli.login_path and mycli.login_path_as_host:
        prompt_host = mycli.login_path
    elif sqlexecute.host is not None:
        prompt_host = sqlexecute.host
    else:
        prompt_host = DEFAULT_HOST
    short_prompt_host, _, _ = prompt_host.partition('.')
    if re.match(r'^[\d\.]+$', short_prompt_host):
        short_prompt_host = prompt_host
    now = datetime.now()
    species_name = sqlexecute.server_info.species.name if sqlexecute.server_info and sqlexecute.server_info.species else 'MySQL'
    strings = string.split('\\\\')
    is_html = strings[0].startswith('\\<html>')
    strings = [x.replace('\\u', maybe_html_escape(sqlexecute.user or '(none)', is_html)) for x in strings]
    strings = [x.replace('\\h', maybe_html_escape(prompt_host or '(none)', is_html)) for x in strings]
    strings = [x.replace('\\H', maybe_html_escape(short_prompt_host or '(none)', is_html)) for x in strings]
    strings = [x.replace('\\d', maybe_html_escape(sqlexecute.dbname or '(none)', is_html)) for x in strings]
    strings = [x.replace('\\t', maybe_html_escape(species_name, is_html)) for x in strings]
    strings = [x.replace('\\n', '\n') for x in strings]
    strings = [x.replace('\\D', maybe_html_escape(now.strftime('%a %b %d %H:%M:%S %Y'), is_html)) for x in strings]
    strings = [x.replace('\\m', maybe_html_escape(now.strftime('%M'), is_html)) for x in strings]
    strings = [x.replace('\\P', maybe_html_escape(now.strftime('%p'), is_html)) for x in strings]
    strings = [x.replace('\\R', maybe_html_escape(now.strftime('%H'), is_html)) for x in strings]
    strings = [x.replace('\\r', maybe_html_escape(now.strftime('%I'), is_html)) for x in strings]
    strings = [x.replace('\\s', maybe_html_escape(now.strftime('%S'), is_html)) for x in strings]
    strings = [x.replace('\\p', maybe_html_escape(str(sqlexecute.port), is_html)) for x in strings]
    strings = [
        x.replace('\\j', maybe_html_escape(os.path.basename(sqlexecute.socket or '(none)').replace('\\', '/'), is_html)) for x in strings
    ]
    strings = [x.replace('\\J', maybe_html_escape((sqlexecute.socket or '(none)').replace('\\', '/'), is_html)) for x in strings]
    strings = [
        x.replace('\\k', maybe_html_escape(os.path.basename(sqlexecute.socket or str(sqlexecute.port)).replace('\\', '/'), is_html))
        for x in strings
    ]
    strings = [
        x.replace('\\K', maybe_html_escape((sqlexecute.socket or str(sqlexecute.port)).replace('\\', '/'), is_html)) for x in strings
    ]
    strings = [x.replace('\\A', maybe_html_escape(mycli.dsn_alias or '(none)', is_html)) for x in strings]
    strings = [x.replace('\\_', ' ') for x in strings]

    checker_string = ' '.join(strings)
    if hasattr(sqlexecute, 'conn') and sqlexecute.conn is not None:
        if '\\y' in checker_string:
            with sqlexecute.conn.cursor() as cur:
                strings = [x.replace('\\y', maybe_html_escape(str(get_uptime(cur)) or '(none)', is_html)) for x in strings]
        if '\\Y' in checker_string:
            with sqlexecute.conn.cursor() as cur:
                strings = [x.replace('\\Y', maybe_html_escape(format_uptime(str(get_uptime(cur))) or '(none)', is_html)) for x in strings]
    else:
        strings = [x.replace('\\y', '(none)') for x in strings]
        strings = [x.replace('\\Y', '(none)') for x in strings]

    if hasattr(sqlexecute, 'conn') and sqlexecute.conn is not None:
        if '\\T' in checker_string:
            with sqlexecute.conn.cursor() as cur:
                strings = [x.replace('\\T', maybe_html_escape(get_ssl_version(cur) or '(none)', is_html)) for x in strings]
    else:
        strings = [x.replace('\\T', '(none)') for x in strings]

    if hasattr(sqlexecute, 'conn') and sqlexecute.conn is not None:
        if '\\w' in checker_string:
            with sqlexecute.conn.cursor() as cur:
                strings = [x.replace('\\w', maybe_html_escape(str(get_warning_count(cur) or '(none)'), is_html)) for x in strings]
    else:
        strings = [x.replace('\\w', '(none)') for x in strings]
    if hasattr(sqlexecute, 'conn') and sqlexecute.conn is not None:
        if '\\W' in checker_string:
            with sqlexecute.conn.cursor() as cur:
                strings = [x.replace('\\W', maybe_html_escape(str(get_warning_count(cur) or ''), is_html)) for x in strings]
    else:
        strings = [x.replace('\\W', '') for x in strings]

    if is_html:
        strings[0] = strings[0].removeprefix('\\<html>')
        strings[-1] = strings[-1].removesuffix('\\</html>')
    elif '\\x1b' in checker_string:
        strings = [x.replace('\\x1b', '\x1b') for x in strings]

    strings = [re.sub(r'\\(.)', r'(unknown prompt format string: \\\1)', x) for x in strings]

    string = '\\'.join(strings)

    if is_html:
        try:
            formatted_string = to_formatted_text(HTML(string))
        except (ExpatError, ValueError):
            formatted_string = to_formatted_text(HTML('(cannot parse HTML prompt string)'))
    else:
        formatted_string = to_formatted_text(ANSI(string))

    return formatted_string


def _get_prompt_message(
    mycli: 'MyCli',
    app: prompt_toolkit.application.application.Application,
) -> FormattedText:
    if app.current_buffer.text:
        return mycli.last_prompt_message

    prompt = render_prompt_string(mycli, mycli.prompt_format, app.render_counter)
    prompt_plain = to_plain_text(prompt)
    if mycli.prompt_format == mycli.default_prompt and len(prompt_plain) > mycli.max_len_prompt:
        prompt = render_prompt_string(mycli, mycli.default_prompt_splitln, app.render_counter)
        prompt_plain = to_plain_text(prompt)
        mycli.prompt_lines = prompt_plain.count('\n') + 1
    if not mycli.prompt_lines:
        mycli.prompt_lines = prompt_plain.count('\n') + 1

    mycli.last_prompt_message = prompt
    return mycli.last_prompt_message


def _get_continuation(
    mycli: 'MyCli',
    width: int,
    _two: int,
    _three: int,
) -> AnyFormattedText:
    if mycli.multiline_continuation_char == '':
        continuation = ''
    elif mycli.multiline_continuation_char:
        left_padding = width - len(mycli.multiline_continuation_char)
        continuation = ' ' * max((left_padding - 1), 0) + mycli.multiline_continuation_char + ' '
    else:
        continuation = ' '
    return [('class:continuation', continuation)]


def _output_results(
    mycli: 'MyCli',
    state: ReplState,
    results: Generator[SQLResult],
    start: float,
) -> None:
    sqlexecute = mycli.sqlexecute
    assert sqlexecute is not None

    result_count = 0
    watch_count = 0
    for result in results:
        mycli.logger.debug('preamble: %r', result.preamble)
        mycli.logger.debug('header: %r', result.header)
        mycli.logger.debug('rows: %r', result.rows)
        mycli.logger.debug('status: %r', result.status)
        mycli.logger.debug('command: %r', result.command)
        threshold = 1000
        if result.command is not None and result.command['name'] == 'watch':
            if watch_count > 0:
                try:
                    watch_seconds = float(result.command['seconds'])
                    start += watch_seconds
                except ValueError as e:
                    mycli.echo(f'Invalid watch sleep time provided ({e}).', err=True, fg='red')
                    sys.exit(1)
            else:
                watch_count += 1

        if is_select(result.status_plain) and isinstance(result.rows, Cursor) and result.rows.rowcount > threshold:
            mycli.echo(
                f'The result set has more than {threshold} rows.',
                fg='red',
            )
            if not confirm('Do you want to continue?'):
                mycli.echo('Aborted!', err=True, fg='red')
                break

        if mycli.auto_vertical_output:
            if mycli.prompt_session is not None:
                max_width = mycli.prompt_session.output.get_size().columns
            else:
                max_width = DEFAULT_WIDTH
        else:
            max_width = None

        formatted = mycli.format_sqlresult(
            result,
            is_expanded=special.is_expanded_output(),
            is_redirected=special.is_redirected(),
            null_string=mycli.null_string,
            numeric_alignment=mycli.numeric_alignment,
            binary_display=mycli.binary_display,
            max_width=max_width,
        )

        duration = time.time() - start
        try:
            if result_count > 0:
                mycli.echo('')
            try:
                mycli.output(formatted, result)
            except KeyboardInterrupt:
                pass
            if mycli.beep_after_seconds > 0 and duration >= mycli.beep_after_seconds:
                assert mycli.prompt_session is not None
                mycli.prompt_session.output.bell()
            if special.is_timing_enabled():
                mycli.output_timing(f'Time: {duration:0.03f}s')
        except KeyboardInterrupt:
            pass

        start = time.time()
        result_count += 1
        state.mutating = state.mutating or is_mutating(result.status_plain)

        if special.is_show_warnings_enabled() and isinstance(result.rows, Cursor) and result.rows.warning_count > 0:
            warnings = sqlexecute.run('SHOW WARNINGS')
            warnings_duration = time.time() - start
            saw_warning = False
            for warning in warnings:
                saw_warning = True
                formatted = mycli.format_sqlresult(
                    warning,
                    is_expanded=special.is_expanded_output(),
                    is_redirected=special.is_redirected(),
                    null_string=mycli.null_string,
                    numeric_alignment=mycli.numeric_alignment,
                    binary_display=mycli.binary_display,
                    max_width=max_width,
                    is_warnings_style=True,
                )
                mycli.echo('')
                mycli.output(formatted, warning, is_warnings_style=True)

            if saw_warning and special.is_timing_enabled():
                mycli.output_timing(f'Time: {warnings_duration:0.03f}s', is_warnings_style=True)


def _keepalive_hook(
    mycli: 'MyCli',
    _context: Any,
) -> None:
    if mycli.keepalive_ticks is None:
        return
    if mycli.keepalive_ticks < 1:
        return

    mycli._keepalive_counter += 1
    if mycli._keepalive_counter > mycli.keepalive_ticks:
        mycli._keepalive_counter = 0
        mycli.logger.debug('keepalive ping')
        try:
            assert mycli.sqlexecute is not None
            assert mycli.sqlexecute.conn is not None
            mycli.sqlexecute.conn.ping(reconnect=False)
        except Exception as e:
            mycli.logger.debug('keepalive ping error %r', e)


def _build_prompt_session(
    mycli: 'MyCli',
    state: ReplState,
    history: FileHistoryWithTimestamp | None,
    key_bindings: KeyBindings,
) -> None:
    if mycli.toolbar_format.lower() == 'none':
        get_toolbar_tokens = None
    else:
        get_toolbar_tokens = create_toolbar_tokens_func(
            mycli,
            lambda: state.iterations == 0,
            mycli.toolbar_format,
            partial(get_custom_toolbar, mycli),
        )

    if mycli.wider_completion_menu:
        complete_style = CompleteStyle.MULTI_COLUMN
    else:
        complete_style = CompleteStyle.COLUMN

    with mycli._completer_lock:
        if mycli.key_bindings == 'vi':
            editing_mode = EditingMode.VI
        else:
            editing_mode = EditingMode.EMACS

        mycli.prompt_session = PromptSession(
            color_depth=ColorDepth.DEPTH_24_BIT if 'truecolor' in os.getenv('COLORTERM', '').lower() else None,
            lexer=PygmentsLexer(MyCliLexer),
            reserve_space_for_menu=mycli.get_reserved_space(),
            prompt_continuation=lambda width, two, three: _get_continuation(mycli, width, two, three),
            bottom_toolbar=get_toolbar_tokens,
            complete_style=complete_style,
            input_processors=[
                ConditionalProcessor(
                    processor=HighlightMatchingBracketProcessor(chars='[](){}'),
                    filter=has_focus(DEFAULT_BUFFER) & ~is_done,
                )
            ],
            tempfile_suffix='.sql',
            completer=DynamicCompleter(lambda: mycli.completer),
            complete_in_thread=True,
            history=history,
            auto_suggest=ThreadedAutoSuggest(AutoSuggestFromHistory()),
            complete_while_typing=complete_while_typing_filter,
            multiline=cli_is_multiline(mycli),
            style=style_factory_ptoolkit(mycli.syntax_style, mycli.cli_style),
            include_default_pygments_style=False,
            key_bindings=key_bindings,
            enable_open_in_editor=True,
            enable_system_prompt=True,
            enable_suspend=True,
            editing_mode=editing_mode,
            search_ignore_case=True,
        )

        if mycli.key_bindings == 'vi':
            mycli.prompt_session.app.ttimeoutlen = mycli.vi_ttimeoutlen
        else:
            mycli.prompt_session.app.ttimeoutlen = mycli.emacs_ttimeoutlen


def _one_iteration(
    mycli: 'MyCli',
    state: ReplState,
    text: str | None = None,
) -> None:
    sqlexecute = mycli.sqlexecute
    assert sqlexecute is not None

    inputhook = partial(_keepalive_hook, mycli) if mycli.keepalive_ticks and mycli.keepalive_ticks >= 1 else None

    if text is None:
        try:
            assert mycli.prompt_session is not None
            loaded_message_fn = partial(_get_prompt_message, mycli, mycli.prompt_session.app)
            text = mycli.prompt_session.prompt(
                inputhook=inputhook,
                message=loaded_message_fn,
            )
        except KeyboardInterrupt:
            return

        special.set_expanded_output(False)
        special.set_forced_horizontal_output(False)

        try:
            text = handle_editor_command(
                mycli,
                text,
                inputhook,
                loaded_message_fn,
            )
        except RuntimeError as e:
            mycli.logger.error('sql: %r, error: %r', text, e)
            mycli.logger.error('traceback: %r', traceback.format_exc())
            mycli.echo(str(e), err=True, fg='red')
            return

        try:
            if handle_clip_command(mycli, text):
                return
        except RuntimeError as e:
            mycli.logger.error('sql: %r, error: %r', text, e)
            mycli.logger.error('traceback: %r', traceback.format_exc())
            mycli.echo(str(e), err=True, fg='red')
            return

        while special.is_llm_command(text):
            start = time.time()
            try:
                assert sqlexecute.conn is not None
                cur = sqlexecute.conn.cursor()
                context, sql, duration = special.handle_llm(
                    text,
                    cur,
                    sqlexecute.dbname or '',
                    mycli.llm_prompt_field_truncate,
                    mycli.llm_prompt_section_truncate,
                )
                if context:
                    click.echo('LLM Response:')
                    click.echo(context)
                    click.echo('---')
                if special.is_timing_enabled():
                    mycli.output_timing(f'Time: {duration:0.03f}s')
                assert mycli.prompt_session is not None
                text = mycli.prompt_session.prompt(
                    default=sql or '',
                    inputhook=inputhook,
                    message=loaded_message_fn,
                )
            except KeyboardInterrupt:
                return
            except special.FinishIteration as e:
                if e.results:
                    _output_results(mycli, state, e.results, start)
                return
            except RuntimeError as e:
                mycli.logger.error('sql: %r, error: %r', text, e)
                mycli.logger.error('traceback: %r', traceback.format_exc())
                mycli.echo(str(e), err=True, fg='red')
                return

    text = text.strip()
    if not text:
        return

    if is_redirect_command(text):
        sql_part, command_part, file_operator_part, file_part = get_redirect_components(text)
        text = sql_part or ''
        try:
            special.set_redirect(command_part, file_operator_part, file_part)
        except (FileNotFoundError, OSError, RuntimeError) as e:
            mycli.logger.error('sql: %r, error: %r', text, e)
            mycli.logger.error('traceback: %r', traceback.format_exc())
            mycli.echo(str(e), err=True, fg='red')
            return

    if mycli.sandbox_mode and not is_sandbox_allowed(text):
        mycli.echo(
            "ERROR 1820: You must reset your password using ALTER USER or SET PASSWORD before executing this statement.",
            err=True,
            fg='red',
        )
        return

    if mycli.destructive_warning:
        destroy = confirm_destructive_query(mycli.destructive_keywords, text)
        if destroy is None:
            pass
        elif destroy is True:
            mycli.echo('Your call!')
        else:
            mycli.echo('Wise choice!')
            return

    successful = False
    try:
        mycli.logger.debug('sql: %r', text)
        special.write_tee(mycli.last_prompt_message, nl=False)
        special.write_tee(text)
        mycli.log_query(text)

        start = time.time()
        results = sqlexecute.run(text)
        mycli.main_formatter.query = text
        mycli.redirect_formatter.query = text
        successful = True
        _output_results(mycli, state, results, start)
        special.unset_once_if_written(mycli.post_redirect_command)
        special.flush_pipe_once_if_written(mycli.post_redirect_command)
    except pymysql.err.InterfaceError:
        if not mycli.reconnect():
            return
        _one_iteration(mycli, state, text)
        return
    except EOFError as e:
        raise e
    except KeyboardInterrupt:
        connection_id_to_kill = sqlexecute.connection_id or 0
        if connection_id_to_kill > 0:
            mycli.logger.debug('connection id to kill: %r', connection_id_to_kill)
            try:
                sqlexecute.connect()
                for kill_result in sqlexecute.run(f'kill {connection_id_to_kill}'):
                    status_str = str(kill_result.status_plain).lower()
                    if status_str.find('ok') > -1:
                        mycli.logger.debug('cancelled query, connection id: %r, sql: %r', connection_id_to_kill, text)
                        mycli.echo(f'Cancelled query id: {connection_id_to_kill}', err=True, fg='blue')
                    else:
                        mycli.logger.debug(
                            'Failed to confirm query cancellation, connection id: %r, sql: %r',
                            connection_id_to_kill,
                            text,
                        )
                        mycli.echo(f'Failed to confirm query cancellation, id: {connection_id_to_kill}', err=True, fg='red')
            except Exception as e2:
                mycli.echo(f'Encountered error while cancelling query: {e2}', err=True, fg='red')
        else:
            mycli.logger.debug('Did not get a connection id, skip cancelling query')
            mycli.echo('Did not get a connection id, skip cancelling query', err=True, fg='red')
    except NotImplementedError:
        mycli.echo('Not Yet Implemented.', fg='yellow')
    except pymysql.OperationalError as e1:
        mycli.logger.debug('Exception: %r', e1)
        if e1.args[0] == ER_MUST_CHANGE_PASSWORD:
            mycli.sandbox_mode = True
            mycli.echo(
                "ERROR 1820: You must reset your password using ALTER USER or SET PASSWORD before executing this statement.",
                err=True,
                fg='red',
            )
        elif e1.args[0] in (2003, 2006, 2013):
            if not mycli.reconnect():
                return
            _one_iteration(mycli, state, text)
            return
        else:
            mycli.logger.error('sql: %r, error: %r', text, e1)
            mycli.logger.error('traceback: %r', traceback.format_exc())
            mycli.echo(str(e1), err=True, fg='red')
    except Exception as e:
        mycli.logger.error('sql: %r, error: %r', text, e)
        mycli.logger.error('traceback: %r', traceback.format_exc())
        mycli.echo(str(e), err=True, fg='red')
    else:
        if mycli.sandbox_mode and is_password_change(text):
            new_password = extract_new_password(text)
            if new_password is not None:
                sqlexecute.password = new_password
            try:
                sqlexecute.connect()
                mycli.sandbox_mode = False
                mycli.echo("Password changed successfully. Reconnected.", err=True, fg='green')
                mycli.refresh_completions()
            except Exception as e:
                mycli.sandbox_mode = False
                mycli.echo(
                    f"Password changed but reconnection failed: {e}\nPlease restart mycli with your new password.",
                    err=True,
                    fg='yellow',
                )

        if is_dropping_database(text, sqlexecute.dbname):
            sqlexecute.dbname = None
            sqlexecute.connect()

        if need_completion_refresh(text):
            mycli.refresh_completions(reset=need_completion_reset(text))
    finally:
        if mycli.logfile is False:
            mycli.echo('Warning: This query was not logged.', err=True, fg='red')

    query = Query(text, successful, state.mutating)
    mycli.query_history.append(query)


def _contributors_picker() -> str:
    lines: str = ""

    try:
        with resources.files(mycli_package).joinpath("AUTHORS").open('r') as f:
            lines += f.read()
    except FileNotFoundError:
        pass

    contents = []
    for line in lines.split("\n"):
        if m := re.match(r"^ *\* (.*)", line):
            contents.append(m.group(1))
    return random.choice(contents) if contents else 'our contributors'


def _sponsors_picker() -> str:
    lines: str = ""

    try:
        with resources.files(mycli_package).joinpath("SPONSORS").open('r') as f:
            lines += f.read()
    except FileNotFoundError:
        pass

    contents = []
    for line in lines.split("\n"):
        if m := re.match(r"^ *\* (.*)", line):
            contents.append(m.group(1))
    return random.choice(contents) if contents else 'our sponsors'


def _tips_picker() -> str:
    tips = []

    try:
        with resources.files(mycli_package).joinpath('TIPS').open('r') as f:
            for line in f:
                if line.startswith("#"):
                    continue
                if tip := line.strip():
                    tips.append(tip)
    except FileNotFoundError:
        pass

    return random.choice(tips) if tips else r'\? or "help" for help!'


def main_repl(mycli: 'MyCli') -> None:
    sqlexecute = mycli.sqlexecute
    assert sqlexecute is not None
    state = ReplState()

    mycli.configure_pager()
    if mycli.smart_completion and not mycli.sandbox_mode:
        mycli.refresh_completions()

    history = _create_history(mycli)
    key_bindings = mycli_bindings(mycli)
    _show_startup_banner(mycli, sqlexecute)
    _build_prompt_session(mycli, state, history, key_bindings)
    set_all_external_titles(mycli)

    try:
        while True:
            _one_iteration(mycli, state)
            state.iterations += 1
    except EOFError:
        special.close_tee()
        if mycli.verbosity >= 0:
            mycli.echo('Goodbye!')



from functools import partial
import logging
import webbrowser

import prompt_toolkit
from prompt_toolkit.application.current import get_app
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import (
    Condition,
    completion_is_selected,
    control_is_searchable,
    emacs_mode,
    vi_mode,
)
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.named_commands import register as ptoolkit_register
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.selection import SelectionType

from mycli.constants import DOCS_URL
from mycli.packages import key_binding_utils
from mycli.packages.ptoolkit.fzf import search_history
from mycli.packages.ptoolkit.utils import safe_invalidate_display

_logger = logging.getLogger(__name__)


@Condition
def ctrl_d_condition() -> bool:
    """Ctrl-D exit binding is only active when the buffer is empty."""
    app = get_app()
    return not app.current_buffer.text


@Condition
def in_completion() -> bool:
    app = get_app()
    return bool(app.current_buffer.complete_state)


def print_f1_help():
    app = get_app()
    app.print_text('\n')
    app.print_text([
        ('', 'Inline help — type "'),
        ('bold', 'help'),
        ('', '" or "'),
        ('bold', r'\?'),
        ('', '"\n'),
    ])
    app.print_text([
        ('', 'Docs index — '),
        ('bold', DOCS_URL),
        ('', '\n'),
    ])
    app.print_text('\n')


@ptoolkit_register("edit-and-execute-command")
def edit_and_execute(event: KeyPressEvent) -> None:
    """Different from the prompt-toolkit default, we want to have a choice not
    to execute a query after editing, hence validate_and_handle=False."""
    buff = event.current_buffer
    buff.open_in_editor(validate_and_handle=False)


def mycli_bindings(mycli) -> KeyBindings:
    """Custom key bindings for mycli."""
    kb = KeyBindings()

    @kb.add('f1')
    def _(event: KeyPressEvent) -> None:
        """Open browser to documentation index."""
        _logger.debug('Detected F1 key.')
        webbrowser.open_new_tab(DOCS_URL)
        prompt_toolkit.application.run_in_terminal(print_f1_help)
        safe_invalidate_display(event.app)

    @kb.add('escape', '[', 'P')
    def _(event: KeyPressEvent) -> None:
        """Open browser to documentation index."""
        _logger.debug("Detected alternate F1 key sequence.")
        webbrowser.open_new_tab(DOCS_URL)
        prompt_toolkit.application.run_in_terminal(print_f1_help)
        safe_invalidate_display(event.app)

    @kb.add("f2")
    def _(_event: KeyPressEvent) -> None:
        """Enable/Disable SmartCompletion Mode."""
        _logger.debug("Detected F2 key.")
        mycli.completer.smart_completion = not mycli.completer.smart_completion

    @kb.add('escape', '[', 'Q')
    def _(_event: KeyPressEvent) -> None:
        """Enable/Disable SmartCompletion Mode."""
        _logger.debug("Detected alternate F2 key sequence.")
        mycli.completer.smart_completion = not mycli.completer.smart_completion

    @kb.add("f3")
    def _(_event: KeyPressEvent) -> None:
        """Enable/Disable Multiline Mode."""
        _logger.debug("Detected F3 key.")
        mycli.multi_line = not mycli.multi_line

    @kb.add('escape', '[', 'R')
    def _(_event: KeyPressEvent) -> None:
        """Enable/Disable Multiline Mode."""
        _logger.debug('Detected alternate F3 key sequence.')
        mycli.multi_line = not mycli.multi_line

    @kb.add("f4")
    def _(event: KeyPressEvent) -> None:
        """Toggle between Vi and Emacs mode."""
        _logger.debug("Detected F4 key.")
        if mycli.key_bindings == "vi":
            event.app.editing_mode = EditingMode.EMACS
            mycli.key_bindings = "emacs"
            event.app.ttimeoutlen = mycli.emacs_ttimeoutlen
        else:
            event.app.editing_mode = EditingMode.VI
            mycli.key_bindings = "vi"
            event.app.ttimeoutlen = mycli.vi_ttimeoutlen

    @kb.add('escape', '[', 'S')
    def _(event: KeyPressEvent) -> None:
        """Toggle between Vi and Emacs mode."""
        _logger.debug('Detected alternate F4 key sequence.')
        if mycli.key_bindings == 'vi':
            event.app.editing_mode = EditingMode.EMACS
            mycli.key_bindings = 'emacs'
            event.app.ttimeoutlen = mycli.emacs_ttimeoutlen
        else:
            event.app.editing_mode = EditingMode.VI
            mycli.key_bindings = 'vi'
            event.app.ttimeoutlen = mycli.vi_ttimeoutlen

    @kb.add("tab")
    def _(event: KeyPressEvent) -> None:
        """Complete action at cursor."""
        _logger.debug("Detected <Tab> key.")
        b = event.app.current_buffer

        behaviors = mycli.config['keys'].as_list('tab')

        if 'toolkit_default' in behaviors:
            if b.complete_state:
                b.complete_next()
            else:
                b.start_completion(select_first=True)

        if b.complete_state:
            if 'advance' in behaviors:
                b.complete_next()
            elif 'cancel' in behaviors:
                b.cancel_completion()
            return

        if 'advancing_summon' in behaviors:
            b.start_completion(select_first=True)
        elif 'prefixing_summon' in behaviors:
            b.start_completion(insert_common_part=True)
        elif 'summon' in behaviors:
            b.start_completion(select_first=False)

    @kb.add("escape", eager=vi_mode, filter=in_completion)
    def _(event: KeyPressEvent) -> None:
        """Cancel completion menu.

        There will be a lag when canceling Escape due to the processing of
        Alt- keystrokes as Escape- sequences.

        There will be no lag when using control-g to cancel."""
        event.app.current_buffer.cancel_completion()

    @kb.add("c-space")
    def _(event: KeyPressEvent) -> None:
        """
        Complete action at cursor.

        By default, if the autocompletion menu is not showing, display it with the
        appropriate completions for the context.

        If the menu is showing, select the next completion.
        """
        _logger.debug("Detected <C-Space> key.")

        b = event.app.current_buffer

        behaviors = mycli.config['keys'].as_list('control_space')

        if 'toolkit_default' in behaviors:
            if b.text:
                b.start_selection(selection_type=SelectionType.CHARACTERS)
            return

        if b.complete_state:
            if 'advance' in behaviors:
                b.complete_next()
            elif 'cancel' in behaviors:
                b.cancel_completion()
            return

        if 'advancing_summon' in behaviors:
            b.start_completion(select_first=True)
        elif 'prefixing_summon' in behaviors:
            b.start_completion(insert_common_part=True)
        elif 'summon' in behaviors:
            b.start_completion(select_first=False)

    @kb.add("c-x", "p", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Prettify and indent current statement, usually into multiple lines.

        Only accepts buffers containing single SQL statements.
        """
        _logger.debug("Detected <C-x p>/> key.")

        b = event.app.current_buffer
        if b.text:
            b.transform_region(0, len(b.text), partial(key_binding_utils.handle_prettify_binding, mycli))

    @kb.add("c-x", "u", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Unprettify and dedent current statement, usually into one line.

        Only accepts buffers containing single SQL statements.
        """
        _logger.debug("Detected <C-x u>/< key.")

        b = event.app.current_buffer
        if b.text:
            b.transform_region(0, len(b.text), partial(key_binding_utils.handle_unprettify_binding, mycli))

    @kb.add("c-o", "d", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Insert the current date.
        """
        _logger.debug("Detected <C-o d> key.")

        event.app.current_buffer.insert_text(key_binding_utils.server_date(mycli.sqlexecute))

    @kb.add("c-o", "c-d", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Insert the quoted current date.
        """
        _logger.debug("Detected <C-o C-d> key.")

        event.app.current_buffer.insert_text(key_binding_utils.server_date(mycli.sqlexecute, quoted=True))

    @kb.add("c-o", "t", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Insert the current datetime.
        """
        _logger.debug("Detected <C-o t> key.")

        event.app.current_buffer.insert_text(key_binding_utils.server_datetime(mycli.sqlexecute))

    @kb.add("c-o", "c-t", filter=emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """
        Insert the quoted current datetime.
        """
        _logger.debug("Detected <C-o C-t> key.")

        event.app.current_buffer.insert_text(key_binding_utils.server_datetime(mycli.sqlexecute, quoted=True))

    @kb.add("c-r", filter=control_is_searchable)
    def _(event: KeyPressEvent) -> None:
        """Search history using fzf or reverse incremental search."""
        _logger.debug("Detected <C-r> key.")
        mode = mycli.config.get('keys', {}).get('control_r', 'auto')
        if mode == 'reverse_isearch':
            search_history(event, incremental=True)
        else:
            search_history(
                event,
                highlight_preview=mycli.highlight_preview,
                highlight_style=mycli.syntax_style,
            )

    @kb.add("escape", "r", filter=control_is_searchable & emacs_mode)
    def _(event: KeyPressEvent) -> None:
        """Search history using fzf when available."""
        _logger.debug("Detected <alt-r> key.")
        search_history(
            event,
            highlight_preview=mycli.highlight_preview,
            highlight_style=mycli.syntax_style,
        )

    @kb.add('c-d', filter=ctrl_d_condition)
    def _(event: KeyPressEvent) -> None:
        """Exit mycli or ignore keypress."""
        _logger.debug('Detected <C-d> key on empty line.')
        mode = mycli.config.get('keys', {}).get('control_d', 'exit')
        if mode == 'exit':
            event.app.exit(exception=EOFError, style='class:exiting')
        else:
            event.app.output.bell()

    @kb.add("enter", filter=completion_is_selected)
    def _(event: KeyPressEvent) -> None:
        """Makes the enter key work as the tab key only when showing the menu.

        In other words, don't execute query when enter is pressed in
        the completion dropdown menu, instead close the dropdown menu
        (accept current selection).

        """
        _logger.debug("Detected enter key.")

        event.current_buffer.complete_state = None
        b = event.app.current_buffer
        b.complete_state = None

    @kb.add("escape", "enter")
    def _(event: KeyPressEvent) -> None:
        """Introduces a line break in multi-line mode, or dispatches the
        command in single-line mode."""
        _logger.debug("Detected alt-enter key.")
        if mycli.multi_line:
            event.app.current_buffer.validate_and_handle()
        else:
            event.app.current_buffer.insert_text("\n")

    return kb



from __future__ import annotations

from collections.abc import Generator, Iterable
import logging
import os
import re
from typing import TYPE_CHECKING, Any, cast

import click

from mycli.main_modes.repl import set_all_external_titles
from mycli.packages import special
from mycli.packages.filepaths import dir_path_exists
from mycli.packages.interactive_utils import confirm_destructive_query
from mycli.packages.special.main import ArgType, SpecialCommandAlias
from mycli.packages.sqlresult import SQLResult
from mycli.sqlexecute import SQLExecute


class ClientCommandsMixin:
    if TYPE_CHECKING:
        main_formatter: Any
        redirect_formatter: Any
        sqlexecute: Any
        destructive_warning: bool
        destructive_keywords: Any
        config: Any
        prompt_format: str

        def refresh_completions(self, reset: bool = False) -> list[SQLResult]: ...
        def reconnect(self, database: str = '') -> bool: ...
        def echo(self, *args: Any, **kwargs: Any) -> None: ...

    def register_special_commands(self) -> None:
        special.register_special_command(
            self.change_db,
            "use",
            "use <database>",
            "Change to a new database.",
            aliases=[SpecialCommandAlias("\\u", case_sensitive=False)],
        )
        special.register_special_command(
            self.manual_reconnect,
            "connect",
            "connect [database]",
            "Reconnect to the server, optionally switching databases.",
            case_sensitive=True,
            aliases=[SpecialCommandAlias("\\r", case_sensitive=True)],
        )
        special.register_special_command(
            self.refresh_completions,
            "rehash",
            "rehash",
            "Refresh auto-completions.",
            arg_type=ArgType.NO_QUERY,
            aliases=[SpecialCommandAlias("\\#", case_sensitive=False)],
        )
        special.register_special_command(
            self.change_table_format,
            "tableformat",
            "tableformat <format>",
            "Change the table format used to output interactive results.",
            case_sensitive=True,
            aliases=[SpecialCommandAlias("\\T", case_sensitive=True)],
        )
        special.register_special_command(
            self.change_redirect_format,
            "redirectformat",
            "redirectformat <format>",
            "Change the table format used to output redirected results.",
            case_sensitive=True,
            aliases=[SpecialCommandAlias("\\Tr", case_sensitive=True)],
        )
        special.register_special_command(
            self.execute_from_file,
            "source",
            "source <filename>",
            "Execute queries from a file.",
            aliases=[SpecialCommandAlias("\\.", case_sensitive=False)],
        )
        special.register_special_command(
            self.change_prompt_format,
            "prompt",
            "prompt <string>",
            "Change prompt format.",
            case_sensitive=True,
            aliases=[SpecialCommandAlias("\\R", case_sensitive=True)],
        )

    def manual_reconnect(self, arg: str = "", **_) -> Generator[SQLResult, None, None]:
        """
        Interactive method to use for the \r command, so that the utility method
        may be cleanly used elsewhere.
        """
        if not self.reconnect(database=arg):
            yield SQLResult(status="Not connected")
        elif not arg or arg == '``':
            yield SQLResult()
        else:
            yield self.change_db(arg).send(None)

    def change_table_format(self, arg: str, **_) -> Generator[SQLResult, None, None]:
        try:
            self.main_formatter.format_name = arg
            yield SQLResult(status=f"Changed table format to {arg}")
        except ValueError:
            msg = f"Table format {arg} not recognized. Allowed formats:"
            for table_type in self.main_formatter.supported_formats:
                msg += f"\n\t{table_type}"
            yield SQLResult(status=msg)

    def change_redirect_format(self, arg: str, **_) -> Generator[SQLResult, None, None]:
        try:
            self.redirect_formatter.format_name = arg
            yield SQLResult(status=f"Changed redirect format to {arg}")
        except ValueError:
            msg = f"Redirect format {arg} not recognized. Allowed formats:"
            for table_type in self.redirect_formatter.supported_formats:
                msg += f"\n\t{table_type}"
            yield SQLResult(status=msg)

    def change_db(self, arg: str, **_) -> Generator[SQLResult, None, None]:
        if arg.startswith("`") and arg.endswith("`"):
            arg = re.sub(r"^`(.*)`$", r"\1", arg)
            arg = re.sub(r"``", r"`", arg)

        if not arg:
            click.secho("No database selected", err=True, fg="red")
            return

        assert isinstance(self.sqlexecute, SQLExecute)

        if self.sqlexecute.dbname == arg:
            msg = f'You are already connected to database "{self.sqlexecute.dbname}" as user "{self.sqlexecute.user}"'
        else:
            self.sqlexecute.change_db(arg)
            msg = f'You are now connected to database "{self.sqlexecute.dbname}" as user "{self.sqlexecute.user}"'

        # todo: this jump back to repl.py is a sign that separation is incomplete.
        # also: it should not be needed.  Don't titles update on every new prompt?
        set_all_external_titles(cast(Any, self))

        yield SQLResult(status=msg)

    def execute_from_file(self, arg: str, **_) -> Iterable[SQLResult]:
        if not arg:
            message = "Missing required argument: filename."
            return [SQLResult(status=message)]
        try:
            with open(os.path.expanduser(arg)) as f:
                query = f.read()
        except IOError as e:
            return [SQLResult(status=str(e))]

        if self.destructive_warning and confirm_destructive_query(self.destructive_keywords, query) is False:
            message = "Wise choice. Command execution stopped."
            return [SQLResult(status=message)]

        assert isinstance(self.sqlexecute, SQLExecute)
        return self.sqlexecute.run(query)

    def change_prompt_format(self, arg: str, **_) -> list[SQLResult]:
        """
        Change the prompt format.
        """
        if not arg:
            message = "Missing required argument, format."
            return [SQLResult(status=message)]

        self.prompt_format = arg
        return [SQLResult(status=f"Changed prompt format to {arg}")]

    def initialize_logging(self) -> None:
        log_file = os.path.expanduser(self.config["main"]["log_file"])
        log_level = self.config["main"]["log_level"]

        level_map = {
            "CRITICAL": logging.CRITICAL,
            "ERROR": logging.ERROR,
            "WARNING": logging.WARNING,
            "INFO": logging.INFO,
            "DEBUG": logging.DEBUG,
        }

        # Disable logging if value is NONE by switching to a no-op handler
        # Set log level to a high value so it doesn't even waste cycles getting called.
        if log_level.upper() == "NONE":
            handler: logging.Handler = logging.NullHandler()
            log_level = "CRITICAL"
        elif dir_path_exists(log_file):
            handler = logging.FileHandler(log_file)
        else:
            self.echo(f'Error: Unable to open the log file "{log_file}".', err=True, fg="red")
            return

        formatter = logging.Formatter("%(asctime)s (%(process)d/%(threadName)s) %(name)s %(levelname)s - %(message)s")

        handler.setFormatter(formatter)

        root_logger = logging.getLogger("mycli")
        root_logger.addHandler(handler)
        root_logger.setLevel(level_map[log_level.upper()])

        logging.captureWarnings(True)

        root_logger.debug("Initializing mycli logging.")
        root_logger.debug("Log file %r.", log_file)



from __future__ import annotations

from collections import defaultdict
import re
from typing import TYPE_CHECKING, Any

from configobj import ConfigObj

from mycli.config import str_to_bool, strip_matching_quotes

if TYPE_CHECKING:
    from mycli.client import MyCli


def normalize_ssl_mode(config: ConfigObj) -> tuple[str | None, str | None]:
    ssl_mode = config['main'].get('ssl_mode', None) or config['connection'].get('default_ssl_mode', None)
    if ssl_mode not in ('auto', 'on', 'off', None):
        return None, f'Invalid config option provided for ssl_mode ({ssl_mode}); ignoring.'
    return ssl_mode, None


def ensure_my_cnf_sections(my_cnf: ConfigObj) -> None:
    if not my_cnf.get('client'):
        my_cnf['client'] = {}
    if not my_cnf.get('mysqld'):
        my_cnf['mysqld'] = {}


def configure_prompt_state(
    mycli: MyCli,
    config: ConfigObj,
    prompt: str | None,
    prompt_cnf: str | None,
    toolbar_format: str | None,
) -> None:
    mycli.prompt_format = prompt or prompt_cnf or config['main']['prompt'] or mycli.default_prompt
    mycli.prompt_lines = 0
    mycli.multiline_continuation_char = config['main']['prompt_continuation']
    mycli.toolbar_format = toolbar_format or config['main']['toolbar']
    mycli.terminal_tab_title_format = config['main']['terminal_tab_title']
    mycli.terminal_window_title_format = config['main']['terminal_window_title']
    mycli.multiplex_window_title_format = config['main']['multiplex_window_title']
    mycli.multiplex_pane_title_format = config['main']['multiplex_pane_title']


def destructive_keywords_from_config(config: ConfigObj) -> list[str]:
    keywords = config['main'].get('destructive_keywords', 'DROP SHUTDOWN DELETE TRUNCATE ALTER UPDATE')
    return [keyword for keyword in keywords.split(' ') if keyword]


def llm_prompt_truncation(config: ConfigObj) -> tuple[int, int]:
    if 'llm' in config and re.match(r'^\d+$', config['llm'].get('prompt_field_truncate', '')):
        field_truncate = int(config['llm'].get('prompt_field_truncate'))
    else:
        field_truncate = 0
    if 'llm' in config and re.match(r'^\d+$', config['llm'].get('prompt_section_truncate', '')):
        section_truncate = int(config['llm'].get('prompt_section_truncate'))
    else:
        section_truncate = 0
    return field_truncate, section_truncate


class AppStateMixin:
    defaults_suffix: str | None
    login_path: str | None

    def read_my_cnf(self, cnf: ConfigObj, keys: list[str]) -> dict[str, Any]:
        sections = ['client', 'mysqld']
        key_transformations = {
            'mysqld': {
                'socket': 'default_socket',
                'port': 'default_port',
                'user': 'default_user',
            },
        }

        if self.login_path and self.login_path != 'client':
            sections.append(self.login_path)

        if self.defaults_suffix:
            sections.extend([sect + self.defaults_suffix for sect in sections])

        configuration: dict[str, Any] = defaultdict(lambda: None)
        for key in keys:
            for section in cnf:
                if section not in sections or key not in cnf[section]:
                    continue
                new_key = key_transformations.get(section, {}).get(key) or key
                configuration[new_key] = strip_matching_quotes(cnf[section][key])

        return configuration

    def merge_ssl_with_cnf(self, ssl: dict[str, Any], cnf: dict[str, Any]) -> dict[str, Any]:
        merged = {}
        merged.update(ssl)
        prefix = 'ssl-'
        for key, value in cnf.items():
            if not key.startswith(prefix):
                continue
            if value is None:
                continue
            if key == 'ssl-verify-server-cert':
                merged['check_hostname'] = str_to_bool(value)
            else:
                merged[key[len(prefix) :]] = value

        return merged



from typing import Callable

from prompt_toolkit.application import get_app
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.formatted_text import AnyFormattedText, to_formatted_text
from prompt_toolkit.key_binding.vi_state import InputMode

from mycli.packages import special


def create_toolbar_tokens_func(
    mycli,
    show_initial_toolbar_help: Callable[[], bool],
    format_string: str | None,
    get_custom_toolbar: Callable[[str], AnyFormattedText],
) -> Callable[[], list[tuple[str, str]]]:
    """Return a function that generates the toolbar tokens."""

    def get_toolbar_tokens() -> list[tuple[str, str]]:
        divider = ('class:bottom-toolbar', ' │ ')

        result = [("class:bottom-toolbar", "[Tab] Complete")]
        dynamic = []

        result.append(divider)
        result.append(("class:bottom-toolbar", "[F1] Help"))

        if mycli.completer.smart_completion:
            result.append(divider)
            result.append(("class:bottom-toolbar", "[F2] Smart-complete:"))
            result.append(("class:bottom-toolbar.on", "ON "))
        else:
            result.append(divider)
            result.append(("class:bottom-toolbar", "[F2] Smart-complete:"))
            result.append(("class:bottom-toolbar.off", "OFF"))

        if mycli.multi_line:
            result.append(divider)
            result.append(("class:bottom-toolbar", "[F3] Multiline:"))
            result.append(("class:bottom-toolbar.on", "ON "))
        else:
            result.append(divider)
            result.append(("class:bottom-toolbar", "[F3] Multiline:"))
            result.append(("class:bottom-toolbar.off", "OFF"))

        if mycli.prompt_session.editing_mode == EditingMode.VI:
            result.append(divider)
            result.append(("class:bottom-toolbar", "Vi:"))
            result.append(("class:bottom-toolbar.on", _get_vi_mode()))

        if mycli.toolbar_error_message:
            dynamic.append(divider)
            dynamic.append(("class:bottom-toolbar.transaction.failed", mycli.toolbar_error_message))
            mycli.toolbar_error_message = None

        if mycli.multi_line:
            delimiter = special.get_current_delimiter()
            if delimiter != ';' or show_initial_toolbar_help():
                dynamic.append(divider)
                dynamic.append(('class:bottom-toolbar', '"'))
                dynamic.append(('class:bottom-toolbar.on', delimiter))
                dynamic.append(('class:bottom-toolbar', '" ends a statement'))

        if show_initial_toolbar_help():
            dynamic.append(divider)
            dynamic.append(("class:bottom-toolbar", "right-arrow accepts full-line suggestion"))

        if mycli.completion_refresher.is_refreshing():
            dynamic.append(divider)
            dynamic.append(("class:bottom-toolbar", "Refreshing completions…"))

        schema_prefetcher = getattr(mycli, 'schema_prefetcher', None)
        if schema_prefetcher is not None and schema_prefetcher.is_prefetching():
            dynamic.append(divider)
            dynamic.append(("class:bottom-toolbar", "Prefetching schemas…"))

        if format_string and format_string != r'\B':
            if format_string.startswith(r'\B'):
                amended_format = format_string[2:]
                result.extend(dynamic)
                dynamic = []
                result.append(('class:bottom-toolbar', '\n'))
            else:
                amended_format = format_string
                result = []
            formatted = to_formatted_text(get_custom_toolbar(amended_format), style='class:bottom-toolbar')
            result.extend([*formatted])  # coerce to list for mypy

        result.extend(dynamic)
        return result

    return get_toolbar_tokens


def _get_vi_mode() -> str:
    """Get the current vi mode for display."""
    return {
        InputMode.INSERT: "I",
        InputMode.NAVIGATION: "N",
        InputMode.REPLACE: "R",
        InputMode.REPLACE_SINGLE: "R",
        InputMode.INSERT_MULTIPLE: "M",
    }[get_app().vi_state.input_mode]



# type: ignore

from collections.abc import Iterator
import multiprocessing
import os
import platform
import signal
import time
from types import SimpleNamespace
from typing import Any, Callable, Literal, cast

from packaging.version import Version
from prompt_toolkit.formatted_text import (
    ANSI,
)
import pygments
import pymysql
import pytest

from mycli import main
from mycli.constants import (
    DEFAULT_CHARSET,
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_USER,
    TEST_DATABASE,
)
import mycli.output
from mycli.packages import special
from mycli.packages.sqlresult import SQLResult

DATABASE = TEST_DATABASE
PASSWORD = os.getenv("PYTEST_PASSWORD")
USER = os.getenv("PYTEST_USER", DEFAULT_USER)
HOST = os.getenv("PYTEST_HOST", DEFAULT_HOST)
PORT = int(os.getenv("PYTEST_PORT", DEFAULT_PORT))
CHARACTER_SET = os.getenv("PYTEST_CHARSET", DEFAULT_CHARSET)
SSH_USER = os.getenv("PYTEST_SSH_USER", None)
SSH_HOST = os.getenv("PYTEST_SSH_HOST", None)
SSH_PORT = int(os.getenv("PYTEST_SSH_PORT", "22"))
TEMPFILE_PREFIX = 'mycli_test_suite_'

PYGMENTS_VERSION = Version(pygments.__version__)


def pygments_below(version: str) -> bool:
    return PYGMENTS_VERSION < Version(version)


class DummyLogger:
    def __init__(self) -> None:
        self.debug_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []
        self.error_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []
        self.warning_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []

    def debug(self, *args: Any, **kwargs: Any) -> None:
        self.debug_calls.append((args, kwargs))

    def error(self, *args: Any, **kwargs: Any) -> None:
        self.error_calls.append((args, kwargs))

    def warning(self, *args: Any, **kwargs: Any) -> None:
        self.warning_calls.append((args, kwargs))


class DummyFormatter:
    def __init__(self, format_name: str = 'ascii') -> None:
        self.format_name = format_name
        self.query = ''
        self.supported_formats = ['ascii', 'csv', 'tsv', 'vertical']
        self._output_formats = {
            'ascii': SimpleNamespace(formatter_args={'missing_value': mycli.output.DEFAULT_MISSING_VALUE}),
            'csv': SimpleNamespace(formatter_args={'missing_value': mycli.output.DEFAULT_MISSING_VALUE}),
            'tsv': SimpleNamespace(formatter_args={'missing_value': mycli.output.DEFAULT_MISSING_VALUE}),
            'vertical': SimpleNamespace(formatter_args={'missing_value': mycli.output.DEFAULT_MISSING_VALUE}),
        }
        self.calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []

    def format_output(self, rows: Any, header: Any, format_name: str | None = None, **kwargs: Any) -> list[str] | str:
        self.calls.append(((rows, header, format_name), kwargs))
        if format_name == 'vertical':
            return ['vertical output']
        return ['plain output']


class ReusableLock:
    def __init__(self, on_enter: Callable[[], Any] | None = None) -> None:
        self.on_enter = on_enter

    def __enter__(self) -> 'ReusableLock':
        if self.on_enter is not None:
            self.on_enter()
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> Literal[False]:
        return False


class FakeCursorBase:
    def __init__(
        self,
        rows: list[tuple[Any, ...]] | None = None,
        rowcount: int = 0,
        description: list[tuple[Any, ...]] | None = None,
        warning_count: int = 0,
    ) -> None:
        self._rows = list(rows or [])
        self.rowcount = rowcount
        self.description = description or []
        self.warning_count = warning_count

    def __iter__(self) -> Iterator[tuple[Any, ...]]:
        return iter(self._rows)


class RecordingSQLExecute:
    calls: list[dict[str, Any]] = []
    side_effects: list[Any] = []

    def __init__(self, **kwargs: Any) -> None:
        type(self).calls.append(dict(kwargs))
        if type(self).side_effects:
            effect = type(self).side_effects.pop(0)
            if isinstance(effect, BaseException):
                raise effect
            if callable(effect):
                effect(kwargs)
        self.kwargs = kwargs
        self.dbname = kwargs.get('database')
        self.user = kwargs.get('user')
        self.conn = kwargs.get('conn')
        self.sandbox_mode = False


def make_bare_mycli() -> Any:
    cli = object.__new__(main.MyCli)
    cli.logger = cast(Any, DummyLogger())
    cli.main_formatter = DummyFormatter()
    cli.redirect_formatter = DummyFormatter()
    cli.helpers_style = 'helpers-style'
    cli.helpers_warnings_style = 'helpers-warnings-style'
    cli.ptoolkit_style = cast(Any, 'pt-style')
    cli.syntax_style = 'native'
    cli.cli_style = {}
    cli.null_string = '<null>'
    cli.numeric_alignment = 'right'
    cli.binary_display = None
    cli.show_warnings = False
    cli.query_history = []
    cli.toolbar_error_message = None
    cli.prompt_session = None
    cli.last_prompt_message = ANSI('')
    cli.last_custom_toolbar_message = ANSI('')
    cli.prompt_lines = 0
    cli.prompt_format = main.MyCli.default_prompt
    cli.multiline_continuation_char = '>'
    cli.toolbar_format = 'default'
    cli.destructive_warning = False
    cli.destructive_keywords = ['drop']
    cli.keepalive_ticks = None
    cli._keepalive_counter = 0
    cli.verbosity = -1
    cli.smart_completion = False
    cli.key_bindings = 'emacs'
    cli.auto_vertical_output = False
    cli.wider_completion_menu = False
    cli.explicit_pager = False
    cli._completer_lock = cast(Any, ReusableLock())
    cli.prefetch_schemas_mode = 'never'
    cli.prefetch_schemas_list = []
    cli.schema_prefetcher = cast(
        Any,
        SimpleNamespace(
            stop=lambda: None,
            clear_loaded=lambda: None,
            start_configured=lambda: None,
            is_prefetching=lambda: False,
            prefetch_schema_now=lambda schema: None,
        ),
    )
    cli.terminal_tab_title_format = ''
    cli.terminal_window_title_format = ''
    cli.multiplex_window_title_format = ''
    cli.multiplex_pane_title_format = ''
    cli.dsn_alias = None
    cli.login_path = None
    cli.login_path_as_host = False
    cli.post_redirect_command = None
    cli.logfile = None
    cli.emacs_ttimeoutlen = 1.0
    cli.vi_ttimeoutlen = 1.0
    cli.beep_after_seconds = 0.0
    cli.config = {'main': {'history_file': '~/.mycli-history-testing'}}
    cli.output = lambda *args, **kwargs: None  # type: ignore[assignment]
    cli.echo = lambda *args, **kwargs: None  # type: ignore[assignment]
    cli.log_query = lambda *args, **kwargs: None  # type: ignore[assignment]
    cli.log_output = lambda *args, **kwargs: None  # type: ignore[assignment]
    cli.configure_pager = lambda: None  # type: ignore[assignment]
    cli.refresh_completions = lambda reset=False: [SQLResult(status='refresh')]  # type: ignore[assignment]
    cli.reconnect = lambda database='': False  # type: ignore[assignment]
    return cli


def make_dummy_mycli_class(
    *,
    config: dict[str, Any] | None = None,
    my_cnf: dict[str, Any] | None = None,
    config_without_package_defaults: dict[str, Any] | None = None,
) -> Any:
    class DummyMyCli:
        last_instance: Any = None

        def __init__(self, **kwargs: Any) -> None:
            type(self).last_instance = self
            self.init_kwargs = dict(kwargs)
            self.config = config or {'main': {}, 'alias_dsn': {}}
            self.my_cnf = my_cnf or {'client': {}, 'mysqld': {}}
            self.config_without_package_defaults = config_without_package_defaults or {}
            self.default_keepalive_ticks = 5
            self.ssl_mode = None
            self.logger = DummyLogger()
            self.main_formatter = SimpleNamespace(format_name=None)
            self.destructive_warning = False
            self.destructive_keywords = ['drop']
            self.dsn_alias = None
            self.connect_calls: list[dict[str, Any]] = []
            self.run_query_calls: list[tuple[str, Any, bool]] = []
            self.run_cli_called = False
            self.close_called = False
            self.verbosity = 0

        def connect(self, **kwargs: Any) -> None:
            self.connect_calls.append(dict(kwargs))

        def run_query(self, query: str, checkpoint: Any = None, new_line: bool = True) -> None:
            self.run_query_calls.append((query, checkpoint, new_line))

        def run_cli(self) -> None:
            self.run_cli_called = True

        def close(self) -> None:
            self.close_called = True

    return DummyMyCli


def call_click_entrypoint_direct(cli_args: main.CliArgs) -> None:
    assert main.click_entrypoint.callback is not None
    cast(Any, main.click_entrypoint.callback).__wrapped__(cli_args)


def db_connection(dbname=None):
    conn = pymysql.connect(user=USER, host=HOST, port=PORT, database=dbname, password=PASSWORD, charset=CHARACTER_SET, local_infile=False)
    conn.autocommit = True
    return conn


try:
    db_connection()
    CAN_CONNECT_TO_DB = True
except Exception:
    CAN_CONNECT_TO_DB = False

dbtest = pytest.mark.skipif(not CAN_CONNECT_TO_DB, reason=f"Need a mysql instance at {DEFAULT_HOST} accessible by user '{DEFAULT_USER}'")


def create_db(dbname):
    with db_connection().cursor() as cur:
        try:
            cur.execute(f"DROP DATABASE IF EXISTS {TEST_DATABASE}")
            cur.execute(f"CREATE DATABASE {TEST_DATABASE}")
        except Exception:
            pass


def run(executor, sql, rows_as_list=True):
    """Return string output for the sql to be run."""
    results = []

    for result in executor.run(sql):
        rows = list(result.rows) if (rows_as_list and result.rows) else result.rows
        results.append({
            "preamble": result.preamble,
            "header": result.header,
            "rows": rows,
            "postamble": result.postamble,
            "status": result.status,
            "status_plain": result.status_plain,
        })

    return results


def set_expanded_output(is_expanded):
    """Pass-through for the tests."""
    return special.set_expanded_output(is_expanded)


def is_expanded_output():
    """Pass-through for the tests."""
    return special.is_expanded_output()


def send_ctrl_c_to_pid(pid, wait_seconds):
    """Sends a Ctrl-C like signal to the given `pid` after `wait_seconds`
    seconds."""
    time.sleep(wait_seconds)
    system_name = platform.system()
    if system_name == "Windows":
        os.kill(pid, signal.CTRL_C_EVENT)
    else:
        os.kill(pid, signal.SIGINT)


def send_ctrl_c(wait_seconds):
    """Create a process that sends a Ctrl-C like signal to the current process
    after `wait_seconds` seconds.

    Returns the `multiprocessing.Process` created.

    """
    ctrl_c_process = multiprocessing.Process(target=send_ctrl_c_to_pid, args=(os.getpid(), wait_seconds))
    ctrl_c_process.start()
    return ctrl_c_process



"""A module to import instead of paramiko when it is not available (to avoid
checking for paramiko all over the place).

When paramiko is first invoked, this simply shuts down mycli, telling the
user they either have to install paramiko or should not use SSH features.

"""


class Paramiko:
    def __getattr__(self, name: str) -> None:
        import sys
        from textwrap import dedent

        print(
            dedent("""
            To enable certain SSH features you need to install ssh extras:

                pip install 'mycli[ssh]'

            or

                pip install paramiko sshtunnel

            This is required for the following command-line arguments:

                --list-ssh-config
                --ssh-config-host
                --ssh-host
            """),
            file=sys.stderr,
        )
        sys.exit(1)


paramiko = Paramiko()



# type: ignore

from types import SimpleNamespace

import pytest
import sqlparse
from sqlparse.sql import Identifier, IdentifierList, Token, TokenList
from sqlparse.tokens import DML, Keyword, Punctuation

from mycli.packages import sql_utils
from mycli.packages.sql_utils import (
    extract_columns_from_select,
    extract_from_part,
    extract_table_identifiers,
    extract_tables,
    extract_tables_from_complete_statements,
    find_prev_keyword,
    get_last_select,
    is_destructive,
    is_dropping_database,
    is_mutating,
    is_select,
    is_subselect,
    last_word,
    need_completion_refresh,
    need_completion_reset,
    queries_start_with,
    query_has_where_clause,
    query_is_single_table_update,
    query_starts_with,
)


def test_extract_columns_from_select():
    columns = extract_columns_from_select('SELECT COLUMN_NAME, DATA_TYPE, IS_NULLABLE, COLUMN_DEFAULT FROM INFORMATION_SCHEMA.COLUMNS')
    assert columns == ['COLUMN_NAME', 'DATA_TYPE', 'IS_NULLABLE', 'COLUMN_DEFAULT']


def test_extract_columns_from_select_empty():
    columns = extract_columns_from_select('')
    assert columns == []


def test_extract_columns_from_select_update():
    columns = extract_columns_from_select('UPDATE table SET value = 1 WHERE id = 1')
    assert columns == []


def test_empty_string():
    tables = extract_tables('')
    assert tables == []


def test_simple_select_single_table():
    tables = extract_tables('select * from abc')
    assert tables == [(None, 'abc', None)]


def test_simple_select_single_table_schema_qualified():
    tables = extract_tables('select * from abc.def')
    assert tables == [('abc', 'def', None)]


def test_simple_select_multiple_tables():
    tables = extract_tables('select * from abc, def')
    assert sorted(tables) == [(None, 'abc', None), (None, 'def', None)]


def test_simple_select_multiple_tables_schema_qualified():
    tables = extract_tables('select * from abc.def, ghi.jkl')
    assert sorted(tables) == [('abc', 'def', None), ('ghi', 'jkl', None)]


def test_simple_select_with_cols_single_table():
    tables = extract_tables('select a,b from abc')
    assert tables == [(None, 'abc', None)]


def test_simple_select_with_cols_single_table_schema_qualified():
    tables = extract_tables('select a,b from abc.def')
    assert tables == [('abc', 'def', None)]


def test_simple_select_with_cols_multiple_tables():
    tables = extract_tables('select a,b from abc, def')
    assert sorted(tables) == [(None, 'abc', None), (None, 'def', None)]


def test_simple_select_with_cols_multiple_tables_with_schema():
    tables = extract_tables('select a,b from abc.def, def.ghi')
    assert sorted(tables) == [('abc', 'def', None), ('def', 'ghi', None)]


def test_select_with_hanging_comma_single_table():
    tables = extract_tables('select a, from abc')
    assert tables == [(None, 'abc', None)]


def test_select_with_hanging_comma_multiple_tables():
    tables = extract_tables('select a, from abc, def')
    assert sorted(tables) == [(None, 'abc', None), (None, 'def', None)]


def test_select_with_hanging_period_multiple_tables():
    tables = extract_tables('SELECT t1. FROM tabl1 t1, tabl2 t2')
    assert sorted(tables) == [(None, 'tabl1', 't1'), (None, 'tabl2', 't2')]


def test_simple_insert_single_table():
    tables = extract_tables('insert into abc (id, name) values (1, "def")')

    # sqlparse mistakenly assigns an alias to the table
    # assert tables == [(None, 'abc', None)]
    assert tables == [(None, 'abc', 'abc')]


def test_simple_insert_single_table_schema_qualified():
    tables = extract_tables('insert into abc.def (id, name) values (1, "def")')
    assert tables == [('abc', 'def', None)]


def test_simple_update_table():
    tables = extract_tables('update abc set id = 1')
    assert tables == [(None, 'abc', None)]


def test_simple_update_table_with_schema():
    tables = extract_tables('update abc.def set id = 1')
    assert tables == [('abc', 'def', None)]


def test_join_table():
    tables = extract_tables('SELECT * FROM abc a JOIN def d ON a.id = d.num')
    assert sorted(tables) == [(None, 'abc', 'a'), (None, 'def', 'd')]


def test_join_table_schema_qualified():
    tables = extract_tables('SELECT * FROM abc.def x JOIN ghi.jkl y ON x.id = y.num')
    assert tables == [('abc', 'def', 'x'), ('ghi', 'jkl', 'y')]


def test_join_as_table():
    tables = extract_tables('SELECT * FROM my_table AS m WHERE m.a > 5')
    assert tables == [(None, 'my_table', 'm')]


def test_extract_tables_from_complete_statements():
    tables = extract_tables_from_complete_statements('SELECT * FROM my_table AS m WHERE m.a > 5')
    assert tables == [(None, 'my_table', 'm')]


def test_extract_tables_from_complete_statements_cte():
    tables = extract_tables_from_complete_statements('WITH my_cte (id, num) AS ( SELECT id, COUNT(1) FROM my_table GROUP BY id ) SELECT *')
    assert tables == [(None, 'my_table', None)]


# this would confuse plain extract_tables() per #1122
def test_extract_tables_from_multiple_complete_statements():
    tables = extract_tables_from_complete_statements(r'\T sql-insert; SELECT * FROM my_table AS m WHERE m.a > 5')
    assert tables == [(None, 'my_table', 'm')]


def test_query_starts_with():
    query = 'USE test;'
    assert query_starts_with(query, ('use',)) is True

    query = 'DROP DATABASE test;'
    assert query_starts_with(query, ('use',)) is False


def test_query_starts_with_comment():
    query = '# comment\nUSE test;'
    assert query_starts_with(query, ('use',)) is True


def test_queries_start_with():
    sql = '# comment\nshow databases;use foo;'
    assert queries_start_with(sql, ['show', 'select']) is True
    assert queries_start_with(sql, ['use', 'drop']) is True
    assert queries_start_with(sql, ['delete', 'update']) is False


@pytest.mark.parametrize(
    ('text', 'include', 'expected'),
    [
        ('abc', 'alphanum_underscore', 'abc'),
        (' abc', 'alphanum_underscore', 'abc'),
        ('', 'alphanum_underscore', ''),
        (' ', 'alphanum_underscore', ''),
        ('abc ', 'alphanum_underscore', ''),
        ('abc def', 'alphanum_underscore', 'def'),
        ('abc def ', 'alphanum_underscore', ''),
        ('abc def;', 'alphanum_underscore', ''),
        ('bac $def', 'alphanum_underscore', 'def'),
        ('bac $def', 'most_punctuations', '$def'),
        (r'bac \def', 'most_punctuations', r'\def'),
        (r'bac \def;', 'most_punctuations', r'\def;'),
        ('bac::def', 'most_punctuations', 'def'),
        ('abc:def', 'many_punctuations', 'def'),
        ('abc.def', 'all_punctuations', 'abc.def'),
    ],
)
def test_last_word(text, include, expected):
    assert last_word(text, include=include) == expected


def test_is_subselect_returns_false_for_non_group_token():
    token = sqlparse.parse('foo')[0].tokens[0]
    assert is_subselect(token) is False


def test_is_subselect_returns_false_for_group_without_dml():
    token = sqlparse.parse('(foo)')[0].tokens[0]
    assert is_subselect(token) is False


def test_is_subselect_returns_true_for_group_with_select():
    token = sqlparse.parse('(select 1)')[0].tokens[0]
    assert is_subselect(token) is True


def test_get_last_select_returns_empty_token_list_without_select():
    parsed = sqlparse.parse('update t set x = 1')[0]
    assert list(get_last_select(parsed).flatten()) == []


def test_get_last_select_returns_single_select_statement():
    parsed = sqlparse.parse('select c1')[0]
    tokens = get_last_select(parsed)
    assert ''.join(token.value for token in tokens.flatten()) == 'select c1'


def test_get_last_select_returns_single_select_statement_with_from():
    parsed = sqlparse.parse('select c1 from')[0]
    tokens = get_last_select(parsed)
    assert ''.join(token.value for token in tokens.flatten()) == 'select c1 from'


def test_get_last_select_returns_last_top_level_select():
    parsed = sqlparse.parse('select c1 union select c2')[0]
    tokens = get_last_select(parsed)
    assert ''.join(token.value for token in tokens.flatten()) == 'select c2'


def test_get_last_select_keeps_outer_select_for_nested_subselect():
    parsed = sqlparse.parse('select c1 from (select c2')[0]
    tokens = get_last_select(parsed)
    assert ''.join(token.value for token in tokens.flatten()) == 'select c2'


def token_values(tokens):
    return [token.value for token in tokens if not getattr(token, 'is_whitespace', False)]


# todo: coverage of stop_at_punctuation parameter
def test_extract_from_part_returns_identifier_after_from():
    parsed = sqlparse.parse('select * from abc')[0]
    tokens = extract_from_part(parsed)
    assert token_values(tokens) == ['abc']


def test_extract_from_part_returns_identifier_list():
    parsed = sqlparse.parse('select * from abc, def')[0]
    tokens = extract_from_part(parsed)
    assert token_values(tokens) == ['abc, def']


def test_extract_from_part_handles_multiple_joins_and_skips_on_clause():
    parsed = sqlparse.parse('select * from abc join def on abc.id = def.id join ghi')[0]
    tokens = extract_from_part(parsed)
    assert token_values(tokens) == ['abc', 'join', 'def', 'ghi']


def test_extract_from_part_recurses_into_subselect_and_stops_at_punctuation():
    parsed = sqlparse.parse('select * from (select * from inner_table), outer_table')[0]
    tokens = extract_from_part(parsed)
    assert token_values(tokens) == ['inner_table']


def test_extract_from_part_stops_at_punctuation_when_requested():
    parsed = TokenList([Token(Keyword, 'FROM'), Token(Punctuation, ','), Token(Keyword, 'SELECT')])
    tokens = extract_from_part(parsed, stop_at_punctuation=True)
    assert token_values(tokens) == []


def test_extract_table_identifiers_handles_identifier_list():
    parsed = sqlparse.parse('select * from abc a, def d')[0]
    token_stream = extract_from_part(parsed)
    assert list(extract_table_identifiers(token_stream)) == [
        (None, 'abc', 'a'),
        (None, 'def', 'd'),
    ]


def test_extract_table_identifiers_handles_schema_qualified_identifier():
    parsed = sqlparse.parse('select * from abc.def x')[0]
    token_stream = extract_from_part(parsed)
    assert list(extract_table_identifiers(token_stream)) == [('abc', 'def', 'x')]


def test_extract_table_identifiers_handles_function_tokens():
    parsed = sqlparse.parse('select * from my_func()')[0]
    token_stream = extract_from_part(parsed)
    assert list(extract_table_identifiers(token_stream)) == [(None, 'my_func', 'my_func')]


def test_extract_table_identifiers_skips_identifier_list_entries_without_identifier_methods():
    class BrokenIdentifierList(IdentifierList):
        def get_identifiers(self):
            return [object()]

    assert list(extract_table_identifiers(iter([BrokenIdentifierList([])]))) == []


def test_extract_table_identifiers_uses_name_when_identifier_has_no_real_name():
    class NamelessIdentifier(Identifier):
        def get_real_name(self):
            return None

        def get_parent_name(self):
            return None

        def get_name(self):
            return 'fallback_name'

        def get_alias(self):
            return None

    assert list(extract_table_identifiers(iter([NamelessIdentifier([])]))) == [
        (None, 'fallback_name', 'fallback_name'),
    ]


@pytest.mark.parametrize(
    ('sql', 'expected_keyword', 'expected_text'),
    [
        ('', None, ''),
        ('foo', None, ''),
        ('select * from foo where bar = 1', 'where', 'select * from foo where'),
        ('select * from foo where a = 1 and b = 2', 'where', 'select * from foo where'),
        ('select * from foo where a between 1 and 2', 'where', 'select * from foo where'),
        ('select count(', '(', 'select count('),
    ],
)
def test_find_prev_keyword(sql, expected_keyword, expected_text):
    token, text = find_prev_keyword(sql)
    assert (token.value if token else None) == expected_keyword
    assert text == expected_text


@pytest.mark.parametrize(
    ('sql', 'is_single_table'),
    [
        ('update test set x = 1', True),
        ('update test t set x = 1', True),
        ('update /* inline comment */ test set x = 1', True),
        ('select 1', False),
        ('', False),
        ('update', False),
        ('update test, foo set x = 1', False),
        ('update test join foo on test.id = foo.id set test.x = 1', False),
    ],
)
def test_query_is_single_table_update(sql, is_single_table):
    assert query_is_single_table_update(sql) is is_single_table


def test_extract_columns_from_select_handles_falsey_last_select(monkeypatch):
    monkeypatch.setattr(sql_utils, 'get_last_select', lambda _parsed: [])
    assert extract_columns_from_select('select 1') == []


def test_extract_columns_from_select_handles_single_identifier(monkeypatch):
    class SingleIdentifier(Identifier):
        def get_real_name(self):
            return 'column_name'

    monkeypatch.setattr(
        sql_utils,
        'get_last_select',
        lambda _parsed: TokenList([Token(DML, 'SELECT'), SingleIdentifier([])]),
    )

    assert extract_columns_from_select('select column_name') == ['column_name']


def test_extract_columns_from_select_ignores_unhandled_identifier_list_entries(monkeypatch):
    class WeirdIdentifierList(IdentifierList):
        def get_identifiers(self):
            return [object()]

    monkeypatch.setattr(
        sql_utils,
        'get_last_select',
        lambda _parsed: TokenList([Token(DML, 'SELECT'), WeirdIdentifierList([])]),
    )

    assert extract_columns_from_select('select 1') == []


def test_extract_columns_from_select_stops_at_keyword_before_collecting_columns(monkeypatch):
    monkeypatch.setattr(
        sql_utils,
        'get_last_select',
        lambda _parsed: TokenList([Token(DML, 'SELECT'), Token(Keyword, 'FROM')]),
    )

    assert extract_columns_from_select('select 1') == []


def test_extract_tables_from_complete_statements_returns_empty_for_falsey_rough_parse(monkeypatch):
    monkeypatch.setattr(sql_utils.sqlparse, 'parse', lambda _sql: [])

    assert extract_tables_from_complete_statements('select * from t') == []


def test_extract_tables_from_complete_statements_skips_cte_table_identifiers(monkeypatch):
    class FakeParentSelect:
        def sql(self):
            return 'WITH cte AS (SELECT 1) SELECT * FROM cte'

    class FakeIdentifier:
        parent_select = FakeParentSelect()
        db = ''
        name = 'cte'
        alias = ''

    class FakeStatement:
        def find_all(self, _table_type):
            return [FakeIdentifier()]

    monkeypatch.setattr(sql_utils.sqlparse, 'parse', lambda _sql: ['stmt'])
    monkeypatch.setattr(sql_utils.sqlglot, 'parse_one', lambda *_args, **_kwargs: FakeStatement())

    assert extract_tables_from_complete_statements('with cte as (select 1) select * from cte') == []


def test_query_is_single_table_update_returns_false_when_parse_result_is_empty(monkeypatch):
    monkeypatch.setattr(sql_utils.sqlparse, 'parse', lambda _sql: [])

    assert query_is_single_table_update('update test set x = 1') is False


def test_is_destructive():
    sql = "use test;\nshow databases;\ndrop database foo;"
    assert is_destructive(["drop"], sql) is True


def test_is_destructive_update_with_where_clause():
    sql = "use test;\nshow databases;\nUPDATE test SET x = 1 WHERE id = 1;"
    assert is_destructive(["update"], sql) is False


def test_is_destructive_update_with_where_clause_and_comment():
    sql = "use test;\nshow databases;\nUPDATE /* inline comment */ test SET x = 1 WHERE id = 1;"
    assert is_destructive(["update"], sql) is False


def test_is_destructive_update_multiple_tables_with_where_clause():
    sql = "use test;\nshow databases;\nUPDATE test, foo SET x = 1 WHERE id = 1;"
    assert is_destructive(["update"], sql) is True


def test_is_destructive_update_without_where_clause():
    sql = "use test;\nshow databases;\nUPDATE test SET x = 1;"
    assert is_destructive(["update"], sql) is True


def test_is_destructive_skips_empty_split_queries(monkeypatch):
    monkeypatch.setattr(sql_utils.sqlparse, 'split', lambda _queries: ['', ''])

    assert is_destructive(['drop'], 'ignored') is False


def test_is_destructive_returns_false_when_no_query_matches_keywords() -> None:
    assert is_destructive(['drop'], 'select 1; show databases;') is False


@pytest.mark.parametrize(
    ("sql", "has_where_clause"),
    [
        ("update test set dummy = 1;", False),
        ("update test set dummy = 1 where id = 1);", True),
    ],
)
def test_query_has_where_clause(sql, has_where_clause):
    assert query_has_where_clause(sql) is has_where_clause


@pytest.mark.parametrize(
    ("sql", "dbname", "is_dropping"),
    [
        ("select bar from foo", "foo", False),
        ('drop database "foo";', "`foo`", True),
        ("drop schema foo", "foo", True),
        ("drop schema foo", "bar", False),
        ("drop database bar", "foo", False),
        ("drop database foo", None, False),
        ("drop database foo; create database foo", "foo", False),
        ("drop database foo; create database bar", "foo", True),
        ("select bar from foo; drop database bazz", "foo", False),
        ("select bar from foo; drop database bazz", "bazz", True),
        ("-- dropping database \n drop -- really dropping \n schema abc -- now it is dropped", "abc", True),
    ],
)
def test_is_dropping_database(sql, dbname, is_dropping):
    assert is_dropping_database(sql, dbname) == is_dropping


def test_is_dropping_database_skips_statements_without_enough_keywords():
    assert is_dropping_database('drop foo', 'foo') is False


@pytest.mark.parametrize(
    ('queries', 'expected'),
    [
        ('select 1;', False),
        ('alter table foo add column bar int;', True),
        ('create table foo (id int);', True),
        ('use foo;', True),
        ('\\r foo localhost root', True),
        ('\\u foo', True),
        ('connect foo localhost root', True),
        ('drop table foo;', True),
        ('rename table foo to bar;', True),
    ],
)
def test_need_completion_refresh(queries, expected):
    assert need_completion_refresh(queries) is expected


def test_need_completion_refresh_ignores_queries_that_fail_to_split(monkeypatch):
    class BrokenQuery:
        def split(self):
            raise RuntimeError('broken')

    monkeypatch.setattr(sql_utils.sqlparse, 'split', lambda _queries: [BrokenQuery(), 'select 1;'])

    assert need_completion_refresh('ignored') is False


@pytest.mark.parametrize(
    ('queries', 'expected'),
    [
        ('select 1;', False),
        ('use foo;', True),
        ('\\u foo', True),
        ('\\r', False),
        ('\\r foo localhost root', True),
        ('connect', False),
        ('connect foo localhost root', True),
    ],
)
def test_need_completion_reset(queries, expected):
    assert need_completion_reset(queries) is expected


def test_need_completion_reset_ignores_queries_that_fail_to_split(monkeypatch):
    class BrokenQuery:
        def split(self):
            raise RuntimeError('broken')

    monkeypatch.setattr(sql_utils.sqlparse, 'split', lambda _queries: [BrokenQuery(), 'select 1;'])

    assert need_completion_reset('ignored') is False


def test_classify_sandbox_statement_treats_token_error_as_quit(monkeypatch):
    def raise_token_error(*_args, **_kwargs):
        raise sql_utils.sqlglot.errors.TokenError('bad token')

    monkeypatch.setattr(sql_utils.sqlglot, 'tokenize', raise_token_error)

    assert sql_utils.classify_sandbox_statement('`') == ('quit', None)


def test_classify_sandbox_statement_treats_empty_tokens_as_quit(monkeypatch):
    monkeypatch.setattr(sql_utils.sqlglot, 'tokenize', lambda *_args, **_kwargs: [])

    assert sql_utils.classify_sandbox_statement('ignored') == ('quit', None)


def test_find_password_after_eq_returns_none_for_non_string_token() -> None:
    token_type = sql_utils.sqlglot.tokens.TokenType
    tokens = [
        SimpleNamespace(token_type=token_type.EQ, text='='),
        SimpleNamespace(token_type=token_type.VAR, text='CURRENT_USER'),
    ]

    assert sql_utils._find_password_after_eq(tokens) is None


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ('', ('quit', None)),
        ('  ', ('quit', None)),
        ('quit', ('quit', None)),
        ('exit', ('quit', None)),
        ('\\q', ('quit', None)),
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'new'", ('alter_user', 'new')),
        ('ALTER USER root IDENTIFIED WITH mysql_native_password', ('alter_user', None)),
        ("SET PASSWORD = 'newpass'", ('set_password', 'newpass')),
        ('SELECT 1', (None, None)),
    ],
)
def test_classify_sandbox_statement(text: str, expected: tuple[str | None, str | None]) -> None:
    assert sql_utils.classify_sandbox_statement(text) == expected


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ('', True),
        ('  ', True),
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'new'", True),
        ('alter user root identified by "pw"', True),
        ("SET PASSWORD = 'newpass'", True),
        ("set password = 'newpass'", True),
        ('quit', True),
        ('exit', True),
        ('\\q', True),
        ('SELECT 1', False),
        ('DROP TABLE t', False),
        ('USE mydb', False),
        ('SHOW DATABASES', False),
    ],
)
def test_is_sandbox_allowed(text: str, expected: bool) -> None:
    assert sql_utils.is_sandbox_allowed(text) is expected


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'new'", True),
        ("SET PASSWORD = 'newpass'", True),
        ('SELECT 1', False),
        ('quit', False),
    ],
)
def test_is_password_change(text: str, expected: bool) -> None:
    assert sql_utils.is_password_change(text) is expected


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'newpass'", 'newpass'),
        ("SET PASSWORD = 'secret123'", 'secret123'),
        ("ALTER USER root IDENTIFIED BY 'p@ss w0rd!'", 'p@ss w0rd!'),
        ('ALTER USER root IDENTIFIED WITH mysql_native_password', None),
        ('SELECT 1', None),
    ],
)
def test_extract_new_password(text: str, expected: str | None) -> None:
    assert sql_utils.extract_new_password(text) == expected


@pytest.mark.parametrize(
    ('status_plain', 'expected'),
    [
        (None, False),
        ('', False),
        ('SELECT 1', False),
        ('INSERT 1', True),
        ('update 3', True),
        ('rename table', True),
    ],
)
def test_is_mutating(status_plain, expected):
    assert is_mutating(status_plain) is expected


@pytest.mark.parametrize(
    ('status_plain', 'expected'),
    [
        (None, False),
        ('', False),
        ('SELECT 1', True),
        ('select rows', True),
        ('UPDATE 1', False),
    ],
)
def test_is_select(status_plain, expected):
    assert is_select(status_plain) is expected
